# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from enum import IntEnum


class States(IntEnum):
    M1 = 1
    M2 = 2
    M3 = 3
    M4 = 4
    M5 = 5
    M6 = 6



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import argparse
import sys
import logging
import tlv8
import base64

from homekit.controller import Controller
from homekit.log_support import setup_logging, add_log_arguments
from homekit.model.characteristics import CharacteristicsDecoderLoader
from homekit.controller.tools import AbstractPairing
from homekit.model.characteristics.characteristic_formats import CharacteristicFormats


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit get_characteristic - retrieve values of characteristics '
                                                 'and other information from paired HomeKit accessories.')
    parser.add_argument('-f', action='store', required=True, dest='file', help='File with the pairing data')
    parser.add_argument('-a', action='store', required=True, dest='alias', help='alias for the pairing')
    parser.add_argument('-c', action='append', required=True, dest='characteristics',
                        help='Read characteristics, multiple characteristics can be given by repeating the option')
    parser.add_argument('-m', action='store_true', required=False, dest='meta',
                        help='read out the meta data for the characteristics as well')
    parser.add_argument('-p', action='store_true', required=False, dest='perms',
                        help='read out the permissions for the characteristics as well')
    parser.add_argument('-t', action='store_true', required=False, dest='type',
                        help='read out the types for the characteristics as well')
    parser.add_argument('-e', action='store_true', required=False, dest='events',
                        help='read out the events for the characteristics as well')
    parser.add_argument('-d', action='store_true', required=False, dest='decode',
                        help='If set, try to find a decoder for each characteristic and show the decoded value')
    add_log_arguments(parser)
    return parser.parse_args()


def get_characteristic_decoders(pairing: AbstractPairing) -> dict:
    """
    This function filters characteristics of an accessory for decodable types (currently only TLV characteristics) and
    tries to load the decoder functions for said characteristic's uuid.

    :param pairing: an implementation of `AbstractPairing` (either for IP or BLE)
    :return: a dict of aid/cid to decoder functions
    """
    loaded_decoders = {}
    loader = CharacteristicsDecoderLoader()
    for a in pairing.list_accessories_and_characteristics():
        aid = a['aid']
        for s in a['services']:
            for c in s['characteristics']:
                c_format = c['format']
                # TODO what about CharacteristicFormats.data?
                if c_format not in [CharacteristicFormats.tlv8]:
                    continue
                c_id = c['iid']
                key = '{}.{}'.format(aid, c_id)
                c_type = c['type']
                decoder = loader.load(c_type)
                if decoder is not None:
                    loaded_decoders[key] = decoder

    return loaded_decoders


if __name__ == '__main__':
    args = setup_args_parser()

    setup_logging(args.loglevel)

    controller = Controller()
    controller.load_data(args.file)
    if args.alias not in controller.get_pairings():
        print('"{a}" is no known alias'.format(a=args.alias))
        sys.exit(-1)

    pairing = controller.get_pairings()[args.alias]

    # convert the command line parameters to the required form
    characteristics = [(int(c.split('.')[0]), int(c.split('.')[1])) for c in args.characteristics]

    # get the data
    try:
        data = pairing.get_characteristics(characteristics, include_meta=args.meta, include_perms=args.perms,
                                           include_type=args.type, include_events=args.events)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    if args.decode:
        decoders = get_characteristic_decoders(pairing)
    else:
        decoders = {}

    # print the data
    tmp = {}
    for k in data:
        nk = str(k[0]) + '.' + str(k[1])
        value = data[k]

        if decoders.get(nk):
            try:
                value['value'] = decoders.get(nk)(base64.b64decode(value['value']))
            except Exception as e:
                logging.ERROR('could not decode', e)

        tmp[nk] = value

    print(json.dumps(tmp, indent=4, cls=tlv8.JsonEncoder))



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse

from homekit.controller import Controller
from homekit.log_support import setup_logging, add_log_arguments


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit remove pairing app')
    parser.add_argument('-f', action='store', required=True, dest='file', help='File with the pairing data')
    parser.add_argument('-a', action='store', required=True, dest='alias',
                        help='alias for the accessory')
    parser.add_argument('-i', action='store', required=False, dest='controllerPairingId',
                        help='this pairing ID identifies the controller who should be removed from accessory')
    parser.add_argument('--adapter', action='store', dest='adapter', default='hci0',
                        help='the bluetooth adapter to be used (defaults to hci0)')
    add_log_arguments(parser)
    return parser.parse_args()


if __name__ == '__main__':
    args = setup_args_parser()

    setup_logging(args.loglevel)

    controller = Controller(args.adapter)
    controller.load_data(args.file)
    if args.alias not in controller.get_pairings():
        print('"{a}" is no known alias'.format(a=args.alias))
        exit(-1)

    controller.remove_pairing(args.alias, args.controllerPairingId)
    controller.save_data(args.file)
    print('Pairing for "{a}" was removed.'.format(a=args.alias))



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class _FeatureFlags(object):
    """
    Data taken form table 5-8 Bonjour TXT Record Feature Flags on page 69.
    """

    def __init__(self):
        self._data = {
            0: 'No support for HAP Pairing',
            1: 'Supports HAP Pairing'
        }

    def __getitem__(self, item):
        bit_value = item & 0x01
        if bit_value in self._data:
            return self._data[bit_value]

        raise KeyError('Item {item} not found'.format(item=item))


FeatureFlags = _FeatureFlags()



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    'CharacteristicsTypes', 'CharacteristicPermissions', 'AbstractCharacteristic', 'BatteryLevelCharacteristic',
    'BatteryLevelCharacteristicMixin', 'BrightnessCharacteristic', 'BrightnessCharacteristicMixin',
    'CharacteristicFormats', 'CharacteristicUnits', 'CurrentHeatingCoolingStateCharacteristic',
    'CurrentHeatingCoolingStateCharacteristicMixin', 'CurrentTemperatureCharacteristic',
    'CurrentTemperatureCharacteristicMixin', 'FirmwareRevisionCharacteristic', 'HardwareRevisionCharacteristic',
    'HueCharacteristic', 'HueCharacteristicMixin', 'IdentifyCharacteristic', 'ManufacturerCharacteristic',
    'ModelCharacteristic', 'NameCharacteristic', 'OnCharacteristic', 'OnCharacteristicMixin',
    'OutletInUseCharacteristic', 'OutletInUseCharacteristicMixin', 'SaturationCharacteristic',
    'SaturationCharacteristicMixin', 'SerialNumberCharacteristic', 'TargetHeatingCoolingStateCharacteristic',
    'TargetHeatingCoolingStateCharacteristicMixin', 'TargetTemperatureCharacteristic',
    'TargetTemperatureCharacteristicMixin', 'TemperatureDisplayUnitCharacteristic', 'TemperatureDisplayUnitsMixin',
    'VolumeCharacteristic', 'VolumeCharacteristicMixin', 'CharacteristicsDecoderLoader'
]

import importlib
import logging
import pkgutil

from homekit.model.characteristics.characteristic_permissions import CharacteristicPermissions
from homekit.model.characteristics.characteristic_types import CharacteristicsTypes
from homekit.model.characteristics.characteristic_units import CharacteristicUnits
from homekit.model.characteristics.characteristic_formats import CharacteristicFormats

from homekit.model.characteristics.abstract_characteristic import AbstractCharacteristic
from homekit.model.characteristics.battery_level import BatteryLevelCharacteristic, BatteryLevelCharacteristicMixin
from homekit.model.characteristics.brightness import BrightnessCharacteristicMixin, BrightnessCharacteristic
from homekit.model.characteristics.current_heating_cooling_state import CurrentHeatingCoolingStateCharacteristicMixin, \
    CurrentHeatingCoolingStateCharacteristic
from homekit.model.characteristics.current_temperature import CurrentTemperatureCharacteristicMixin, \
    CurrentTemperatureCharacteristic
from homekit.model.characteristics.firmware_revision import FirmwareRevisionCharacteristic
from homekit.model.characteristics.hardware_revision import HardwareRevisionCharacteristic
from homekit.model.characteristics.hue import HueCharacteristicMixin, HueCharacteristic
from homekit.model.characteristics.identify import IdentifyCharacteristic
from homekit.model.characteristics.manufacturer import ManufacturerCharacteristic
from homekit.model.characteristics.model import ModelCharacteristic
from homekit.model.characteristics.name import NameCharacteristic
from homekit.model.characteristics.on import OnCharacteristicMixin, OnCharacteristic
from homekit.model.characteristics.outlet_in_use import OutletInUseCharacteristic, OutletInUseCharacteristicMixin
from homekit.model.characteristics.saturation import SaturationCharacteristicMixin, SaturationCharacteristic
from homekit.model.characteristics.serialnumber import SerialNumberCharacteristic
from homekit.model.characteristics.target_heating_cooling_state import TargetHeatingCoolingStateCharacteristic, \
    TargetHeatingCoolingStateCharacteristicMixin
from homekit.model.characteristics.target_temperature import TargetTemperatureCharacteristicMixin, \
    TargetTemperatureCharacteristic
from homekit.model.characteristics.temperature_display_unit import TemperatureDisplayUnitsMixin, \
    TemperatureDisplayUnitCharacteristic
from homekit.model.characteristics.volume import VolumeCharacteristic, VolumeCharacteristicMixin


class CharacteristicsDecoderLoader:
    """
    class to dynamically load decoders for tlv8 characteristics.
    """

    def __init__(self):
        self.decoders = {}

    def load(self, char_type: str):
        """
        This function loads a decoder for the specified characteristics:
         - get the name of the characteristic via the given uuid (via `CharacteristicsTypes.get_short()`)
         - load a module from `homekit.model.characteristics` plus the name of the characteristic
         - the module must contain a function `decoder`

        :param char_type: the uuid of the characteristic
        :return: a function that decodes the value of the characteristic into a `tlv8.EntryList`
        """
        characteristic_name = CharacteristicsTypes.get_short(char_type)
        if characteristic_name.startswith('Unknown'):
            mod_name = 'uuid_{}'.format(char_type.replace('-', '_'))
            logging.info('modname %s', mod_name)
        else:
            mod_name = characteristic_name.replace('-', '_')
        if char_type not in self.decoders:

            # try to dynamically load from the standard characteristics by name
            try:
                logging.info('loading module "%s" for type "%s"', mod_name, char_type)
                module = importlib.import_module('homekit.model.characteristics.' + mod_name)
                decoder = getattr(module, 'decoder')
                self.decoders[char_type] = decoder
                return decoder
            except Exception as e:
                logging.info('Error loading decoder: "%s" for type "%s"', e, char_type)

            # try to load from all plugins, it may be a non-standard characteristic with vendor specific data
            try:
                for _, plugin_name, _ in pkgutil.iter_modules():
                    if not plugin_name.startswith('homekit_'):
                        continue
                    logging.info('loading module "%s" for type "%s" from plugin "%s"', mod_name, char_type, plugin_name)
                    module = importlib.import_module('.model.characteristics.' + mod_name, plugin_name)
                    decoder = getattr(module, 'decoder')
                    self.decoders[char_type] = decoder
                    return decoder
            except Exception as e:
                logging.info('Error loading decoder: "%s" for type "%s"', e, char_type)

            return None
        else:
            logging.info('got decoder for %s from cache', char_type)
            return self.decoders[char_type]



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from homekit.exceptions import HttpException


class HttpResponse(object):
    STATE_PRE_STATUS = 0
    STATE_HEADERS = 1
    STATE_BODY = 2
    STATE_DONE = 3

    def __init__(self):
        self._state = HttpResponse.STATE_PRE_STATUS
        self._raw_response = bytearray()
        self._is_ready = False
        self._is_chunked = False
        self._had_empty_chunk = False
        self._content_length = -1
        self.version = None
        self.code = None
        self.reason = None
        self.headers = []
        self.body = bytearray()

    def parse(self, part):
        self._raw_response += part
        pos = self._raw_response.find(b'\r\n')
        while pos != -1:
            line = self._raw_response[:pos]
            self._raw_response = self._raw_response[pos + 2:]
            if self._state == HttpResponse.STATE_PRE_STATUS:
                # parse status line
                line = line.split(b' ', 2)
                if len(line) != 3:
                    raise HttpException('Malformed status line.')
                self.version = line[0].decode()
                self.code = int(line[1])
                self.reason = line[2].decode()
                self._state = HttpResponse.STATE_HEADERS

            elif self._state == HttpResponse.STATE_HEADERS and line == b'':
                # this is the empty line after the headers
                self._state = HttpResponse.STATE_BODY

            elif self._state == HttpResponse.STATE_HEADERS:
                # parse a header line
                line = line.split(b':', 1)
                name = line[0].decode()
                value = line[1].decode().strip()
                if name == 'Transfer-Encoding':
                    if value == 'chunked':
                        self._is_chunked = True
                elif name == 'Content-Length':
                    self._content_length = int(value)
                self.headers.append((name, value))

            elif self._state == HttpResponse.STATE_BODY:
                if self._is_chunked:
                    length = int(line, 16)
                    if length + 2 > len(self._raw_response):
                        self._raw_response = line + b'\r\n' + self._raw_response
                        # the remaining bytes in raw response are not sufficient. bail out and wait for an other call.
                        break
                    if length == 0:
                        self._had_empty_chunk = True
                        self._state = HttpResponse.STATE_DONE
                        self._raw_response = self._raw_response[length + 2:]
                    else:
                        line = self._raw_response[:length]
                        self.body += line
                        self._raw_response = self._raw_response[length + 2:]
                if self._content_length > -1:
                    self.body += self._raw_response
                    self._raw_response = bytearray()
            else:
                raise HttpException('Unknown parser state')

            pos = self._raw_response.find(b'\r\n')

        if self._state == HttpResponse.STATE_BODY and self._content_length > 0:
            remaining = self._content_length - len(self.body)
            self.body += self._raw_response[:remaining]
            self._raw_response = self._raw_response[remaining:]

        if self.is_read_completely():
            # Whatever is left in the buffer is part of the next request
            return self._raw_response

        return bytearray()

    def read(self):
        """
        Returns the body of the response.

        :return: The read body or None if no body content was read yet
        """
        return self.body

    def is_read_completely(self):
        if self._is_chunked:
            return self._had_empty_chunk
        if self.code == 204:
            return True
        if self._content_length != -1:
            return len(self.body) == self._content_length
        if self._state == HttpResponse.STATE_PRE_STATUS or self._state == HttpResponse.STATE_HEADERS:
            return False
        raise HttpException('Could not determine if HTTP data was read completely')

    def get_http_name(self):
        """
        Returns the HTTP name (e.g. HTTP or EVENT).

        :return: The name or None if the status line was not yet read
        """
        if self.version is not None:
            return self.version.split('/')[0]
        return None



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import base64
import binascii
from decimal import Decimal
import struct

from homekit.model.mixin import ToDictMixin
from homekit.model.characteristics import CharacteristicsTypes, CharacteristicFormats, CharacteristicPermissions
from homekit.protocol.statuscodes import HapStatusCodes
from homekit.exceptions import CharacteristicPermissionError, FormatError

def strtobool(val):
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    if val in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    raise ValueError(f"invalid truth value {val!r}")


class AbstractCharacteristic(ToDictMixin):
    def __init__(self, iid: int, characteristic_type: str, characteristic_format: str):
        if type(self) is AbstractCharacteristic:
            raise Exception('AbstractCharacteristic is an abstract class and cannot be instantiated directly')
        self.type = CharacteristicsTypes.get_uuid(characteristic_type)  # page 65, see ServicesTypes
        self.iid = iid  # page 65, unique instance id
        self.perms = [CharacteristicPermissions.paired_read]  # page 65, array of values from CharacteristicPermissions
        self.format = characteristic_format  # page 66, one of CharacteristicsTypes
        self.value = None  # page 65, required but depends on format

        self.ev = None  # boolean, not required, page 65
        self.description = None  # string, not required, page 65
        self.unit = None  # string, not required,page 66, valid values are in CharacteristicUnits
        self.minValue = None  # number, not required, page 66, used if format is int* or float
        self.maxValue = None  # number, not required, page 66, used if format is int* or float
        self.minStep = None  # number, not required, page 66, used if format is int* or float
        self.maxLen = 64  # number, not required, page 66, used if format is string
        self.maxDataLen = 2097152  # number, not required, page 66, used if format is data
        self.valid_values = None  # array, not required, see page 67, all numeric entries are allowed values
        self.valid_values_range = None  # 2 element array, not required, see page 67

        self._set_value_callback = None
        self._get_value_callback = None

    def set_set_value_callback(self, callback):
        self._set_value_callback = callback

    def set_get_value_callback(self, callback):
        self._get_value_callback = callback

    def set_events(self, new_val):
        self.ev = new_val

    def set_value(self, new_val):
        """
        This function sets the value of this characteristic. Permissions are checked first

        :param new_val:
        :raises CharacteristicPermissionError: if the characteristic cannot be written
        """
        if CharacteristicPermissions.paired_write not in self.perms:
            raise CharacteristicPermissionError(HapStatusCodes.CANT_WRITE_READ_ONLY)
        try:
            # convert input to python int if it is any kind of int
            if self.format in [CharacteristicFormats.uint64, CharacteristicFormats.uint32, CharacteristicFormats.uint16,
                               CharacteristicFormats.uint8, CharacteristicFormats.int]:
                new_val = int(new_val)
            # convert input to python float
            if self.format == CharacteristicFormats.float:
                new_val = float(new_val)
            # convert to python bool
            if self.format == CharacteristicFormats.bool:
                new_val = strtobool(str(new_val))
        except ValueError:
            raise FormatError(HapStatusCodes.INVALID_VALUE)

        if self.format in [CharacteristicFormats.uint64, CharacteristicFormats.uint32, CharacteristicFormats.uint16,
                           CharacteristicFormats.uint8, CharacteristicFormats.int, CharacteristicFormats.float]:
            if self.minValue is not None and new_val < self.minValue:
                raise FormatError(HapStatusCodes.INVALID_VALUE)
            if self.maxValue is not None and self.maxValue < new_val:
                raise FormatError(HapStatusCodes.INVALID_VALUE)
            if self.minStep is not None:
                tmp = new_val

                # if minValue is set, the steps count from this on
                if self.minValue is not None:
                    tmp -= self.minValue

                # use Decimal to calculate the module because it has not the precision problem as float...
                if Decimal(str(tmp)) % Decimal(str(self.minStep)) != 0:
                    raise FormatError(HapStatusCodes.INVALID_VALUE)
            if self.valid_values is not None and new_val not in self.valid_values:
                raise FormatError(HapStatusCodes.INVALID_VALUE)
            if self.valid_values_range is not None and not (
                    self.valid_values_range[0] <= new_val <= self.valid_values_range[1]):
                raise FormatError(HapStatusCodes.INVALID_VALUE)

        if self.format == CharacteristicFormats.data:
            try:
                byte_data = base64.decodebytes(new_val.encode())
            except binascii.Error:
                raise FormatError(HapStatusCodes.INVALID_VALUE)
            except Exception:
                raise FormatError(HapStatusCodes.OUT_OF_RESOURCES)
            if self.maxDataLen < len(byte_data):
                raise FormatError(HapStatusCodes.INVALID_VALUE)

        if self.format == CharacteristicFormats.string:
            if len(new_val) > self.maxLen:
                raise FormatError(HapStatusCodes.INVALID_VALUE)

        self.value = new_val
        if self._set_value_callback:
            self._set_value_callback(new_val)

    def set_value_from_ble(self, value):
        if self.format == CharacteristicFormats.bool:
            value = struct.unpack('?', value)[0]
        elif self.format == CharacteristicFormats.uint8:
            value = struct.unpack('B', value)[0]
        elif self.format == CharacteristicFormats.uint16:
            value = struct.unpack('H', value)[0]
        elif self.format == CharacteristicFormats.uint32:
            value = struct.unpack('I', value)[0]
        elif self.format == CharacteristicFormats.uint64:
            value = struct.unpack('Q', value)[0]
        elif self.format == CharacteristicFormats.int:
            value = struct.unpack('i', value)[0]
        elif self.format == CharacteristicFormats.float:
            value = struct.unpack('f', value)[0]
        elif self.format == CharacteristicFormats.string:
            value = value.decode('UTF-8')
        else:
            value = value.hex()

        self.set_value(value)

    def get_value(self):
        """
        This method returns the value of this characteristic. Permissions are checked first, then either the callback
        for getting the values is executed (execution time may vary) or the value is directly returned if not callback
        is given.

        :raises CharacteristicPermissionError: if the characteristic cannot be read
        :return: the value of the characteristic
        """
        if CharacteristicPermissions.paired_read not in self.perms:
            raise CharacteristicPermissionError(HapStatusCodes.CANT_READ_WRITE_ONLY)
        if self._get_value_callback:
            return self._get_value_callback()
        return self.value

    def get_value_for_ble(self):
        value = self.get_value()

        if self.format == CharacteristicFormats.bool:
            try:
                val = strtobool(str(value))
            except ValueError:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=self.format))

            value = struct.pack('?', val)
        elif self.format == CharacteristicFormats.int:
            value = struct.pack('i', int(value))
        elif self.format == CharacteristicFormats.float:
            value = struct.pack('f', float(value))
        elif self.format == CharacteristicFormats.string:
            value = value.encode()

        return value

    def get_meta(self):
        """
        This method returns a dict of meta information for this characteristic. This includes at least the format of
        the characteristic but may contain any other specific attribute.

        :return: a dict
        """
        tmp = {'format': self.format}
        # TODO implement handling of already defined maxLen (upto 256!)
        if self.format == CharacteristicFormats.string:
            tmp['maxLen'] = 64
        # TODO implement handling of other fields! eg maxDataLen
        return tmp

    def to_accessory_and_service_list(self):
        d = {
            'type': self.type,
            'iid': self.iid,
            'perms': self.perms,
            'format': self.format,
        }
        if CharacteristicPermissions.paired_read in self.perms:
            d['value'] = self.value
        if self.ev:
            d['ev'] = self.ev
        if self.description:
            d['description'] = self.description
        if self.unit:
            d['unit'] = self.unit
        if self.minValue:
            d['minValue'] = self.minValue
        if self.maxValue:
            d['maxValue'] = self.maxValue
        if self.minStep:
            d['minStep'] = self.minStep
        if self.maxLen and self.format in [CharacteristicFormats.string]:
            d['maxLen'] = self.maxLen

        return d



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tlv8
from enum import IntEnum
from homekit.model.characteristics.characteristic_types import CharacteristicsTypes

CHARACTERISTIC_ID = CharacteristicsTypes.get_uuid(CharacteristicsTypes.STREAMING_STATUS)


class StreamingStatusKeys(IntEnum):
    """
    Page 215 / Table 9-17
    """
    STATUS = 1


class StreamingStatusValues(IntEnum):
    """
    Page 215 / Table 9-17 Values for Key Status
    """
    AVAILABLE = 0
    IN_USE = 1
    UNAVAILABLE = 2


def decoder(bytes_data):
    return tlv8.decode(bytes_data, {
        StreamingStatusKeys.STATUS: StreamingStatusValues
    })



#!/usr/bin/env python3

#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os.path
import logging
import argparse
import sys
import json
import tlv8
import base64
import importlib
import functools

from homekit import AccessoryServer, Controller
from homekit.accessoryserver import AccessoryRequestHandler
from homekit.model import Accessory
from homekit.model.services import AbstractService, ServicesTypes
from homekit.model.characteristics import AbstractCharacteristic, CharacteristicFormats, CharacteristicsTypes, \
    CharacteristicsDecoderLoader
from homekit.http_impl import HttpStatusCodes
from homekit.controller.tools import AbstractPairing
from homekit.log_support import setup_logging, add_log_arguments

# global containers for the filter functions
get_filters = {}
set_filters = {}


def log_loaded_filter_count():
    """
    logs the numbers of loaded filters.
    """

    def count_filters(filters):
        count = 0
        for accessory_data in filters.items():
            count += len(accessory_data[1].items())
        return count

    logging.info('loaded %i set_filter functions', count_filters(set_filters))
    logging.info('loaded %i get_filter functions', count_filters(get_filters))


def get_filter(aid, cid):
    """
    The get_filter decorator. This adds the decorated function to the global list of get_filters for later usage. The
    original function is also surrounded to return a value in any case. If the decorated function returns None, the
    original value is returned instead.

    :param aid: the accessory's id
    :param cid: the characteristic's id
    :return: a function decorating the filter function
    """

    def decorator_getter(func):
        @functools.wraps(func)
        def decorated(val):
            new_val = func(val)
            if new_val is None:
                return val
            else:
                return new_val

        if aid not in get_filters:
            get_filters[aid] = {}
        if cid in get_filters[aid]:
            raise Exception()
        else:
            get_filters[aid][cid] = func

        return decorated

    return decorator_getter


def set_filter(aid, cid):
    """
    The set_filter decorator. This adds the decorated function to the global list of set_filters for later usage. The
    original function is also surrounded to return a value in any case. If the decorated function returns None, the
    original value is returned instead.

    :param aid: the accessory's id
    :param cid: the characteristic's id
    :return: a function decorating the filter function
    """

    def decorator_setter(func):
        @functools.wraps(func)
        def decorated(val):
            new_val = func(val)
            if new_val is None:
                return val
            else:
                return new_val

        if aid not in set_filters:
            set_filters[aid] = {}
        if cid in set_filters[aid]:
            raise Exception()
        else:
            set_filters[aid][cid] = decorated

        return decorated

    return decorator_setter


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit debug proxy')
    parser.add_argument('-c', '--client-data', action='store', required=True, dest='client_data',
                        default='./client.json', help='JSON file with the pairing data for the accessory')
    parser.add_argument('-a', '--alias', action='store', required=True, dest='alias', help='alias for the pairing')
    parser.add_argument('-s', '--server-data', action='store', required=True, dest='server_data',
                        default='./server.json', help='JSON file with the accessory data to the controller')
    parser.add_argument('-C', '--code', action='store', required=False, dest='code',
                        help='Reference to a python module with filter functions')
    add_log_arguments(parser, 'INFO')
    return parser.parse_args()


class ProxyService(AbstractService):
    """
    Implementation of `AbstractService` to act as a proxy.
    """

    def __init__(self, iid: int, service_type: str):
        AbstractService.__init__(self, service_type, iid)


class ProxyCharacteristic(AbstractCharacteristic):
    """
    Implementation of `AbstractCharacteristic` to act as a proxy.
    """

    def __init__(self, iid: int, characteristic_type: str, characteristic_format: str):
        AbstractCharacteristic.__init__(self, iid, characteristic_type, characteristic_format)


decoder_loader = CharacteristicsDecoderLoader()


def log_transferred_value(text: str, aid: int, characteristic: AbstractCharacteristic, value, filtered_value):
    """
    Logs the transfer of a value between controller and acccessory or vice versa. For characteristics
    of type TLV8, a decoder is used if available else a deep decode is done.

    :param text: a `str` to express which direction of transfer takes place
    :param aid: the accessory id
    :param characteristic: the characteristic for which the transfer takes place
    :param value: the value that was transferred
    """
    iid = int(characteristic.iid)
    debug_value = value
    filtered_debug_value = filtered_value
    characteristic_name = CharacteristicsTypes.get_short(characteristic.type)
    if characteristic.format == CharacteristicFormats.tlv8:
        bytes_value = base64.b64decode(value)
        filtered_bytes_value = base64.b64decode(filtered_value)
        decoder = decoder_loader.load(characteristic.type)
        if decoder:
            try:
                debug_value = tlv8.format_string(decoder(bytes_value))
                filtered_debug_value = tlv8.format_string(decoder(filtered_bytes_value))
            except Exception as e:
                logging.error('problem decoding', e)
        else:
            debug_value = tlv8.format_string(tlv8.deep_decode(bytes_value))
            filtered_debug_value = tlv8.format_string(tlv8.deep_decode(filtered_bytes_value))
    logging.info(
        '%s %s.%s (type %s / %s): \n\toriginal value: %s\n\tfiltered value: %s' % (
            text, aid, iid, characteristic.type, characteristic_name, debug_value, filtered_debug_value))


def generate_set_value_callback(aid: int, characteristic: AbstractCharacteristic, set_filters):
    """
    generate a call back function for the set value use case. This also logs the transferred value.

    :param aid: the id of the accessory
    :param characteristic: the characteristic that is read
    :return: the value as if it would come directly from the proxied device.
    """

    def callback(value):
        iid = int(characteristic.iid)

        # put value through possible filter
        filtered_value = set_filters.get(aid, {}).get(iid, lambda x: x)(value)

        # now log the value
        log_transferred_value('write value to', aid, characteristic, value, filtered_value)

        # set the value on the proxied device
        characteristics = [(aid, iid, filtered_value)]
        pairing.put_characteristics(characteristics)

    return callback


def generate_get_value_callback(aid: int, characteristic: AbstractCharacteristic, get_filters):
    """
    generate a call back function for the get value use case. This also logs the transferred value.

    :param aid: the id of the accessory
    :param characteristic: the characteristic that is read
    :return: the value as if it would come directly from the proxied device.
    """

    def callback():
        iid = int(characteristic.iid)
        characteristics = [(aid, iid)]
        value = pairing.get_characteristics(characteristics)[(int(aid), int(iid))]['value']

        # put value through possible filter
        filtered_value = get_filters.get(aid, {}).get(iid, lambda x: x)(value)

        # now log the value
        log_transferred_value('get value from', aid, characteristic, value, filtered_value)

        # return the value from the proxied device
        return filtered_value

    return callback


def generate_proxy_accessory_request_handler(pairing: AbstractPairing):
    """
    Generate a subclass of `AccessoryRequestHandler` whose method `_post_resource` was over written to proxy all calls
    to the accessory represented by the given pairing. Logs are also done.

    :param pairing: an instance of `AbstractPairing`
    :return: a subclass that can be used as `request_handler_class` for an `AccessoryServer`
    """

    class ProxyAccessoryRequestHandler(AccessoryRequestHandler):

        def _post_resource(self):
            request_body = json.loads(self.body)
            try:
                (content_type, result_body) = pairing.get_resource(request_body)
                logging.info(
                    'post resource with {v} results in {t} with {r} bytes'.format(v=request_body, t=content_type,
                                                                                  r=len(result_body)))
                self.send_response(HttpStatusCodes.OK)
                self.send_header('Content-Type', content_type)
                self.send_header('Content-Length', len(result_body))
                self.end_headers()
                self.wfile.write(result_body)
            except Exception as e:
                logging.error('post resource with {v} results in exception {r}'.format(v=request_body, r=e))
                self.send_error(HttpStatusCodes.INTERNAL_SERVER_ERROR)

    return ProxyAccessoryRequestHandler


def create_proxy(accessories_and_characteristics):
    """
    Create a proxy in front of a set of accessories, services and characteristics. This allows to follow the
    communication of a controller and an accessory (e.g. an iPhone and some HomeKit IP camera).

    :param accessories_and_characteristics: the accessory data as described in the spec on page 73 and following. This
    contains all the data that will be used to create the proxy.
    :return: a list of `Accessory` instances whose services and characteristics are replaced by proxied versions. That
    means characteristic's callback functions for getting and setting values relay those calls to the proxied
    characteristics.
    """
    accessories = []
    logging.info('%<------ creating proxy ------')
    for accessory in accessories_and_characteristics:
        proxy_accessory = Accessory('', '', '', '', '', )
        aid = accessory['aid']
        proxy_accessory.aid = aid
        logging.info('accessory with aid=%s', aid)
        proxy_accessory.services = []
        accessories.append(proxy_accessory)

        for service in accessory['services']:
            service_iid = service['iid']
            service_type = service['type']
            short_type = ServicesTypes.get_short(service_type)
            logging.info('  %i.%i: >%s< (%s)', aid, service_iid, short_type, service_type)

            proxy_service = ProxyService(service_iid, service_type)
            proxy_accessory.add_service(proxy_service)

            for characteristic in service['characteristics']:
                characteristic_iid = characteristic['iid']
                characteristic_type = characteristic['type']
                short_type = CharacteristicsTypes.get_short(characteristic_type)
                characteristic_format = characteristic['format']
                characteristic_value = characteristic.get('value')
                characteristic_perms = characteristic['perms']
                logging.info('    %i.%i: %s >%s< (%s) [%s] %s', aid, characteristic_iid, characteristic_value,
                             short_type, characteristic_type, ','.join(characteristic_perms), characteristic_format)

                proxy_characteristic = ProxyCharacteristic(characteristic_iid, characteristic_type,
                                                           characteristic_format)
                proxy_service.append_characteristic(proxy_characteristic)
                if characteristic_value:
                    proxy_characteristic.value = characteristic_value
                proxy_characteristic.perms = characteristic_perms

                proxy_characteristic.set_set_value_callback(
                    generate_set_value_callback(accessory['aid'], proxy_characteristic, set_filters))
                proxy_characteristic.set_get_value_callback(
                    generate_get_value_callback(accessory['aid'], proxy_characteristic, get_filters))
    logging.info('%<------ finished creating proxy ------')
    return accessories


if __name__ == '__main__':
    args = setup_args_parser()

    setup_logging(args.loglevel)

    client_config_file = os.path.expanduser(args.client_data)
    server_config_file = os.path.expanduser(args.server_data)

    if args.code:
        logging.info('loading filters from "%s"', args.code)
        try:
            module = importlib.import_module(args.code)
            set_filters = getattr(module, 'set_filters')
            get_filters = getattr(module, 'get_filters')
        except Exception as e:
            url = 'https://github.com/jlusiardi/homekit_python/tree/master#filter-functions'
            logging.error('error loading from "%s": %s (see %s for more information)', args.code, e, url)
            sys.exit(-1)
        log_loaded_filter_count()

    controller = Controller()
    try:
        controller.load_data(client_config_file)
    except Exception as e:
        logging.error(e, exc_info=True)
        sys.exit(-1)

    if args.alias not in controller.get_pairings():
        logging.error('"%s" is no known alias', args.alias)
        sys.exit(-1)

    try:
        pairing = controller.get_pairings()[args.alias]
        data = pairing.list_accessories_and_characteristics()
    except Exception as e:
        logging.error(e, exc_info=True)
        sys.exit(-1)

    proxy_accessories = create_proxy(data)

    # create a server and an accessory an run it unless ctrl+c was hit
    try:
        httpd = AccessoryServer(server_config_file, logging.getLogger(),
                                request_handler_class=generate_proxy_accessory_request_handler(pairing))
        for proxy_accessory in proxy_accessories:
            httpd.add_accessory(proxy_accessory)

        httpd.publish_device()
        logging.info('published device and start serving')
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass

    # unpublish the device and shut down
    logging.info('unpublish device')
    httpd.unpublish_device()
    httpd.shutdown()



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tlv8
from enum import IntEnum
from homekit.model.characteristics.characteristic_types import CharacteristicsTypes

CHARACTERISTIC_ID = CharacteristicsTypes.get_uuid(CharacteristicsTypes.SUPPORTED_AUDIO_CONFIGURATION)


class SupportedAudioStreamConfigurationKeys(IntEnum):
    """
    Page 215 / Table 9-18
    """
    AUDIO_CODEC_CONFIGURATION = 1
    COMFORT_NOISE_SUPPORT = 2


class AudioCodecConfigurationKeys(IntEnum):
    """
    Page 215 / Table 9-18
    """
    CODEC_TYPE = 1
    AUDIO_CODEC_PARAMETERS = 2


class AudioCodecTypeValues(IntEnum):
    """
    Page 207 / Table 9-11 Values for key 'Selected Audio Codec type'
    """
    AAC_ELD = 2
    OPUS = 3
    AMR = 5
    AMR_WB = 6


class AudioCodecParametersKeys(IntEnum):
    """
    Page 217 / Table 9-21
    """
    AUDIO_CHANNELS = 1
    BIT_RATE = 2
    SAMPLE_RATE = 3
    RTP_TIME = 4


class BitRateValues(IntEnum):
    """
    Page 217 / Table 9-21 values for key 'Bit-rate'
    """
    VARIABLE = 0
    CONSTANT = 1


class SampleRateValues(IntEnum):
    """
    Page 217 / Table 9-21 values for key 'Sample rate'
    """
    KHZ_8 = 0
    KHZ_16 = 1
    KHZ_24 = 2


class RtpTimeValues(IntEnum):
    """
    Page 217 / Table 9-21 values for key 'RTP Time'
    """
    _20_MS = 20
    _30_MS = 30
    _40_MS = 40
    _60_MS = 60


def decoder(bytes_data):
    return tlv8.decode(bytes_data, {
        SupportedAudioStreamConfigurationKeys.AUDIO_CODEC_CONFIGURATION: {
            AudioCodecConfigurationKeys.CODEC_TYPE: AudioCodecTypeValues,
            AudioCodecConfigurationKeys.AUDIO_CODEC_PARAMETERS: {
                AudioCodecParametersKeys.AUDIO_CHANNELS: tlv8.DataType.UNSIGNED_INTEGER,
                AudioCodecParametersKeys.BIT_RATE: BitRateValues,
                AudioCodecParametersKeys.SAMPLE_RATE: SampleRateValues,
                AudioCodecParametersKeys.RTP_TIME: RtpTimeValues
            }
        },
        SupportedAudioStreamConfigurationKeys.COMFORT_NOISE_SUPPORT: tlv8.DataType.UNSIGNED_INTEGER
    })



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tlv8
from enum import IntEnum
from homekit.model.characteristics.supported_video_stream_configuration import VideoAttributesKeys, \
    VideoCodecParametersKeys, ProfileIdValues, VideoCodecTypeValues, LevelValues, CVOEnabledValue, \
    PacketizationModeValues
from homekit.model.characteristics.supported_audio_configuration import AudioCodecParametersKeys, \
    BitRateValues, SampleRateValues, RtpTimeValues
from homekit.model.characteristics.characteristic_types import CharacteristicsTypes

CHARACTERISTIC_ID = CharacteristicsTypes.get_uuid(CharacteristicsTypes.SELECTED_RTP_STREAM_CONFIGURATION)


class SelectedRtpStreamConfigurationKeys(IntEnum):
    """
    Page 204 / Table 9-7
    """
    SESSION_CONTROL = 1
    SELECTED_VIDEO_PARAMS = 2
    SELECTED_AUDIO_PARAMS = 3


class SessionControlKeys(IntEnum):
    """
    Page 205 / Table 9-8
    """
    SESSION_IDENTIFIER = 1
    COMMAND = 2


class CommandValues(IntEnum):
    """
    Page 205 / Table 9-8 Values for Key Command
    """
    END = 0
    START = 1
    SUSPEND = 2
    RESUME = 3
    RECONFIGURE = 4


class SelectedAudioParametersKeys(IntEnum):
    """
    Page 207 / Table 9-11
    """
    SELECTED_AUDIO_CODEC_TYPE = 1
    SELECTED_AUDIO_CODEC_PARAMETERS = 2
    SELECTED_AUDIO_RTP_PARAMETERS = 3
    COMFORT_NOISE = 4


class SelectedVideoParametersKeys(IntEnum):
    """
    Page 205 / Table 9-9
    """
    SELECTED_VIDEO_CODEC_TYPE = 1
    SELECTED_VIDEO_CODEC_PARAMETERS = 2
    SELECTED_VIDEO_ATTRIBUTES = 3
    SELECTED_VIDEO_RTP_PARAMETERS = 4


class AudioCodecTypeValues(IntEnum):
    """
    Page 207 / Table 9-11 Values for key 'Selected Audio Codec type'
    """
    AAC_ELD = 2
    OPUS = 3
    AMR = 5
    AMR_WB = 6


class VideoRTPParametersKeys(IntEnum):
    """
    Page 206 / Table 9-10
    """
    PAYLOAD_TYPE = 1
    SSRC_FOR_VIDEO = 2
    MAX_BITRATE = 3
    MIN_RTCP = 4
    MAX_MTU = 5


class AudioRtpParametersKeys(IntEnum):
    """
    Page 207 / Table 9-12
    """
    PAYLOAD_TYPE = 1
    SSRC_FOR_AUDIO = 2
    MAX_BITRATE = 3
    MIN_RTCP = 4
    COMFORT_NOISE = 6


def decoder(bytes_data):
    return tlv8.decode(bytes_data, {
        SelectedRtpStreamConfigurationKeys.SESSION_CONTROL: {
            SessionControlKeys.SESSION_IDENTIFIER: tlv8.DataType.BYTES,
            SessionControlKeys.COMMAND: CommandValues,
        },
        SelectedRtpStreamConfigurationKeys.SELECTED_AUDIO_PARAMS: {
            SelectedAudioParametersKeys.SELECTED_AUDIO_CODEC_TYPE: AudioCodecTypeValues,
            SelectedAudioParametersKeys.SELECTED_AUDIO_CODEC_PARAMETERS: {
                AudioCodecParametersKeys.AUDIO_CHANNELS: tlv8.DataType.INTEGER,
                AudioCodecParametersKeys.BIT_RATE: BitRateValues,
                AudioCodecParametersKeys.SAMPLE_RATE: SampleRateValues,
                AudioCodecParametersKeys.RTP_TIME: RtpTimeValues,
            },
            SelectedAudioParametersKeys.SELECTED_AUDIO_RTP_PARAMETERS: {
                AudioRtpParametersKeys.PAYLOAD_TYPE: tlv8.DataType.INTEGER,
                AudioRtpParametersKeys.SSRC_FOR_AUDIO: tlv8.DataType.BYTES,
                AudioRtpParametersKeys.MAX_BITRATE: tlv8.DataType.INTEGER,
                AudioRtpParametersKeys.MIN_RTCP: tlv8.DataType.FLOAT,
                AudioRtpParametersKeys.COMFORT_NOISE: tlv8.DataType.INTEGER,
            },
            SelectedAudioParametersKeys.COMFORT_NOISE: tlv8.DataType.INTEGER,
        },
        SelectedRtpStreamConfigurationKeys.SELECTED_VIDEO_PARAMS: {
            SelectedVideoParametersKeys.SELECTED_VIDEO_CODEC_TYPE: VideoCodecTypeValues,
            SelectedVideoParametersKeys.SELECTED_VIDEO_CODEC_PARAMETERS: {
                VideoCodecParametersKeys.PROFILE_ID: ProfileIdValues,
                VideoCodecParametersKeys.LEVEL: LevelValues,
                VideoCodecParametersKeys.PACKETIZATION_MODE: PacketizationModeValues,
                VideoCodecParametersKeys.CVO_ENABLED: CVOEnabledValue,
            },
            SelectedVideoParametersKeys.SELECTED_VIDEO_ATTRIBUTES: {
                VideoAttributesKeys.IMAGE_WIDTH: tlv8.DataType.INTEGER,
                VideoAttributesKeys.IMAGE_HEIGHT: tlv8.DataType.INTEGER,
                VideoAttributesKeys.FRAME_RATE: tlv8.DataType.INTEGER,
            },
            SelectedVideoParametersKeys.SELECTED_VIDEO_RTP_PARAMETERS: {
                VideoRTPParametersKeys.PAYLOAD_TYPE: tlv8.DataType.INTEGER,
                VideoRTPParametersKeys.SSRC_FOR_VIDEO: tlv8.DataType.BYTES,
                VideoRTPParametersKeys.MAX_BITRATE: tlv8.DataType.INTEGER,
                VideoRTPParametersKeys.MIN_RTCP: tlv8.DataType.FLOAT,
            }
        }
    })



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import argparse
import logging

from homekit.controller import Controller
from homekit.log_support import setup_logging, add_log_arguments


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit identify app - performs identify on given HomeKit device')

    parser.add_argument('--adapter', action='store', dest='adapter', default='hci0',
                        help='the bluetooth adapter to be used (defaults to hci0)')

    group = parser.add_argument_group('Identify unpaired IP', 'use this option to identify an UNPAIRED IP accessory.')
    group.add_argument('-d', action='store', dest='device', help='accessory pairing id')

    group = parser.add_argument_group('Identify unpaired BLE', 'use this option to identify an UNPAIRED BLE accessory.')
    group.add_argument('-m', action='store', dest='mac', help='accessory mac address')

    group = parser.add_argument_group('Identify paired', 'use this option to identify an PAIRED accessory.')
    group.add_argument('-f', action='store', dest='file', help='File with the pairing data')
    group.add_argument('-a', action='store', dest='alias', help='alias for the pairing')
    add_log_arguments(parser)

    parsed_args = parser.parse_args()

    if parsed_args.device and parsed_args.mac is None and parsed_args.file is None and parsed_args.alias is None:
        # unpaired IP
        return parsed_args
    if parsed_args.mac and parsed_args.device is None and parsed_args.file is None and parsed_args.alias is None:
        # unpaired BLE
        return parsed_args
    elif parsed_args.device is None and parsed_args.mac is None and parsed_args.file and parsed_args.alias:
        # paired
        return parsed_args
    else:
        parser.print_help()
        sys.exit(-1)


if __name__ == '__main__':
    args = setup_args_parser()

    setup_logging(args.loglevel)

    controller = Controller(args.adapter)
    if args.device:
        try:
            controller.identify(args.device)
        except Exception as e:
            print(e)
            logging.debug(e, exc_info=True)
            sys.exit(-1)
    elif args.mac:
        try:
            controller.identify_ble(args.mac, args.adapter)
        except Exception as e:
            print(e)
            logging.debug(e, exc_info=True)
            sys.exit(-1)
    else:
        controller.load_data(args.file)
        if args.alias not in controller.get_pairings():
            print('"{a}" is no known alias'.format(a=args.alias))
            exit(-1)

        pairing = controller.get_pairings()[args.alias]
        try:
            pairing.identify()
        except Exception as e:
            print(e)
            logging.debug(e, exc_info=True)
            sys.exit(-1)



# -*- coding: UTF-8 -*-

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from http.client import HTTPConnection


class HomeKitHTTPConnection(HTTPConnection):

    def _send_output(self, message_body=None, encode_chunked=False):
        """
        Acts like the original HTTPConnection._send_output but makes sure the
        headers and the body are sent in one packet. This fixes issues with
        some homekit devices. See for more details:
         * https://github.com/jlusiardi/homekit_python/issues/12
         * https://github.com/jlusiardi/homekit_python/issues/16

        Send the currently buffered request and clear the buffer.

        Appends an extra \\r\\n to the buffer.
        A message_body may be specified, to be appended to the request.
        """
        self._buffer.extend((b"", b""))
        msg = b"\r\n".join(self._buffer)
        del self._buffer[:]

        if message_body is not None:
            msg = msg + message_body

        self.send(msg)



# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from enum import IntEnum


class Methods(IntEnum):
    # Methods (see table 4-4 page 60)
    PairSetup = 1
    PairVerify = 2
    AddPairing = 3
    RemovePairing = 4
    ListPairings = 5

    # Table 6-27 page 116
    kTLVMethod_Resume = 7



#
# Copyright 2019 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from homekit.controller.tools import AbstractPairing


class AdditionalPairing(AbstractPairing):
    """
    This represents an preparation state for an additional pairing (see channel chapter 4.11 Add Pairing page 51).
    """

    def __init__(self, pairing_data):
        """
        Initialize a Pairing by using the data either loaded from file or obtained after calling
        Controller.perform_pairing().

        :param pairing_data:
        """
        self.pairing_data = pairing_data
        self.session = None

    def close(self):
        """
        Close the pairing's communications. This closes the session.
        """
        pass

    def _get_pairing_data(self):
        """
        This method returns the internal pairing data. DO NOT mess around with it.

        :return: a dict containing the data
        """
        return self.pairing_data

    def list_accessories_and_characteristics(self):
        pass

    def list_pairings(self):
        pass

    def get_characteristics(self, characteristics, include_meta=False, include_perms=False, include_type=False,
                            include_events=False):
        pass

    def put_characteristics(self, characteristics, do_conversion=False):
        pass

    def get_events(self, characteristics, callback_fun, max_events=-1, max_seconds=-1):
        pass

    def identify(self):
        pass

    def add_pairing(self, additional_controller_pairing_identifier, ios_device_ltpk, permissions):
        pass



#
# Copyright 2020 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tlv8
from enum import IntEnum
from homekit.model.characteristics.characteristic_types import CharacteristicsTypes

CHARACTERISTIC_ID = CharacteristicsTypes.get_uuid(CharacteristicsTypes.SUPPORTED_VIDEO_STREAM_CONFIGURATION)


class SupportedVideoStreamConfigurationKeys(IntEnum):
    """
    Page 219 / Table 9-25
    """
    VIDEO_CODEC_CONFIGURATION = 1


class VideoCodecConfigurationKeys(IntEnum):
    """
    Page 219 / Table 9-26
    """
    VIDEO_CODEC_TYPE = 1
    VIDEO_CODEC_PARAMETERS = 2
    VIDEO_ATTRIBUTES = 3


class VideoCodecTypeValues(IntEnum):
    """
    Page 219 / Table 9-26 value for 'Video Codec Type'
    """
    H264 = 0


class VideoCodecParametersKeys(IntEnum):
    """
    Page 220 / Table 9-27
    """
    PROFILE_ID = 1
    LEVEL = 2
    PACKETIZATION_MODE = 3
    CVO_ENABLED = 4
    CVO_ID = 5


class ProfileIdValues(IntEnum):
    """
    Page 220 / Table 9-27 Values for key 'ProfileID'
    """
    CONSTRAINED_BASELINE_PROFILE = 0
    MAIN_PROFILE = 1
    HIGH_PROFILE = 2


class LevelValues(IntEnum):
    """
    Page 220 / Table 9-27 Values for key 'Level'
    """
    L_3_1 = 0
    L_3_2 = 1
    L_4 = 2


class CVOEnabledValue(IntEnum):
    """
    Page 220 / Table 9-27 Values for key 'CVO Enabled'
    """
    NOT_SUPPORTED = 0
    SUPPORTED = 1


class PacketizationModeValues(IntEnum):
    """
    Page 220 / Table 9-27 Values for key 'Packetization mode'
    """
    NON_INTERLEAVED = 0


class VideoAttributesKeys(IntEnum):
    """
    Page 220 / Table 9-28
    """
    IMAGE_WIDTH = 1
    IMAGE_HEIGHT = 2
    FRAME_RATE = 3


def decoder(bytes_data):
    return tlv8.decode(bytes_data, {
        SupportedVideoStreamConfigurationKeys.VIDEO_CODEC_CONFIGURATION: {
            VideoCodecConfigurationKeys.VIDEO_CODEC_TYPE: VideoCodecTypeValues,
            VideoCodecConfigurationKeys.VIDEO_CODEC_PARAMETERS: {
                VideoCodecParametersKeys.PROFILE_ID: ProfileIdValues,
                VideoCodecParametersKeys.LEVEL: LevelValues,
                VideoCodecParametersKeys.PACKETIZATION_MODE: PacketizationModeValues,
                VideoCodecParametersKeys.CVO_ENABLED: PacketizationModeValues,
                VideoCodecParametersKeys.CVO_ID: tlv8.DataType.UNSIGNED_INTEGER
            },
            VideoCodecConfigurationKeys.VIDEO_ATTRIBUTES: {
                VideoAttributesKeys.IMAGE_WIDTH: tlv8.DataType.UNSIGNED_INTEGER,
                VideoAttributesKeys.IMAGE_HEIGHT: tlv8.DataType.UNSIGNED_INTEGER,
                VideoAttributesKeys.FRAME_RATE: tlv8.DataType.UNSIGNED_INTEGER,
            },
        }
    })



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import sys
import logging
import re

from homekit.controller import Controller
from homekit.log_support import setup_logging, add_log_arguments


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit IP pairing app')
    parser.add_argument('-d', action='store', required=True, dest='device',
                        help='HomeKit Device ID (use discover to get it)')
    parser.add_argument('-p', action='store', required=False, dest='pin', help='HomeKit configuration code')
    parser.add_argument('-f', action='store', required=True, dest='file', help='HomeKit pairing data file')
    parser.add_argument('-a', action='store', required=True, dest='alias', help='alias for the pairing')
    add_log_arguments(parser)
    return parser.parse_args()


def pin_from_parameter(number):
    def tmp():
        return number
    return tmp


def pin_from_keyboard():
    def tmp():
        read_pin = ''
        while re.match(r'^\d{3}-\d{2}-\d{3}$', read_pin) is None:
            read_pin = input('Enter device pin (XXX-YY-ZZZ): ')
        return read_pin
    return tmp


if __name__ == '__main__':
    args = setup_args_parser()

    setup_logging(args.loglevel)

    controller = Controller()
    try:
        controller.load_data(args.file)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    if args.alias in controller.get_pairings():
        print('"{a}" is a already known alias'.format(a=args.alias))
        sys.exit(-1)

    if args.pin:
        pin_function = pin_from_parameter(args.pin)
    else:
        pin_function = pin_from_keyboard()

    try:
        finish_pairing = controller.start_pairing(args.alias, args.device)
        finish_pairing(pin_function())
        pairing = controller.get_pairings()[args.alias]
        pairing.list_accessories_and_characteristics()
        controller.save_data(args.file)
        print('Pairing for "{a}" was established.'.format(a=args.alias))
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    'AccessoryInformationService', 'BHSLightBulbService', 'FanService', 'LightBulbService', 'ThermostatService',
    'Categories', 'CharacteristicPermissions', 'CharacteristicFormats', 'FeatureFlags', 'Accessory'
]

import json
from homekit.model.mixin import ToDictMixin, get_id
from homekit.model.services import AccessoryInformationService, LightBulbService, FanService, \
    BHSLightBulbService, ThermostatService
from homekit.model.categories import Categories
from homekit.model.characteristics import CharacteristicPermissions, CharacteristicFormats
from homekit.model.feature_flags import FeatureFlags
from homekit.model.characteristics import IdentifyCharacteristic


class Accessory(ToDictMixin):
    def __init__(self, name, manufacturer, model, serial_number, firmware_revision):
        self.aid = get_id()
        self.services = [
            AccessoryInformationService(name, manufacturer, model, serial_number, firmware_revision)
        ]

    def add_service(self, service):
        self.services.append(service)

    def set_identify_callback(self, func):
        """
        Set the callback function for this accessory. This function will be called on paired calls to identify.

        :param func: a function without any parameters and without return type.
        """

        def tmp(x):
            func()

        for service in self.services:
            if isinstance(service, AccessoryInformationService):
                for characteristic in service.characteristics:
                    if isinstance(characteristic, IdentifyCharacteristic):
                        characteristic.set_set_value_callback(tmp)

    def to_accessory_and_service_list(self):
        services_list = []
        for s in self.services:
            services_list.append(s.to_accessory_and_service_list())
        d = {
            'aid': self.aid,
            'services': services_list
        }
        return d


class Accessories(ToDictMixin):
    def __init__(self):
        self.accessories = []

    def add_accessory(self, accessory: Accessory):
        self.accessories.append(accessory)

    def to_accessory_and_service_list(self):
        accessories_list = []
        for a in self.accessories:
            accessories_list.append(a.to_accessory_and_service_list())
        d = {'accessories': accessories_list}
        return json.dumps(d)



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import hashlib
import hkdf
from binascii import hexlify
from nacl.signing import SigningKey
from nacl.encoding import RawEncoder
from nacl.signing import VerifyKey
from nacl.exceptions import BadSignatureError
import logging
import tlv8

from cryptography.hazmat.primitives.asymmetric import x25519
from cryptography.hazmat.primitives import serialization

from homekit.protocol.states import States
from homekit.protocol.methods import Methods
from homekit.protocol.errors import Errors
from homekit.protocol.tlv_types import TlvTypes
from homekit.exceptions import IncorrectPairingIdError, InvalidAuthTagError, InvalidSignatureError, UnavailableError, \
    AuthenticationError, InvalidError, BusyError, MaxTriesError, MaxPeersError, BackoffError

import homekit.exceptions
from homekit.crypto import chacha20_aead_decrypt, chacha20_aead_encrypt, SrpClient


def error_handler(error, stage):
    """
    Transform the various error messages defined in table 4-5 page 60 into exceptions

    :param error: the kind of error
    :param stage: the stage it appeared in
    :return: None
    """
    if error == Errors.Unavailable:
        raise UnavailableError(stage)
    elif error == Errors.Authentication:
        raise AuthenticationError(stage)
    elif error == Errors.Backoff:
        raise BackoffError(stage)
    elif error == Errors.MaxPeers:
        raise MaxPeersError(stage)
    elif error == Errors.MaxTries:
        raise MaxTriesError(stage)
    elif error == Errors.Busy:
        raise BusyError(stage)
    else:
        raise InvalidError(stage)


def create_ip_pair_setup_write(connection):
    def write_http(request, expected):
        body = tlv8.encode(request)
        logging.debug('write message: %s', tlv8.format_string(tlv8.deep_decode(body)))
        connection.putrequest('POST', '/pair-setup', skip_accept_encoding=True)
        connection.putheader('Content-Type', 'application/pairing+tlv8')
        connection.putheader('Content-Length', len(body))
        connection.endheaders(body)
        resp = connection.getresponse()
        response_tlv = tlv8.decode(resp.read(), expected)
        logging.debug('response: %s', tlv8.format_string(response_tlv))
        return response_tlv

    return write_http


def create_ip_pair_verify_write(connection):
    def write_http(request, expected):
        body = tlv8.encode(request)
        logging.debug('write message: %s', tlv8.format_string(tlv8.deep_decode(body)))
        connection.putrequest('POST', '/pair-verify', skip_accept_encoding=True)
        connection.putheader('Content-Type', 'application/pairing+tlv8')
        connection.putheader('Content-Length', len(body))
        connection.endheaders(body)
        resp = connection.getresponse()
        response_tlv = tlv8.decode(resp.read(), expected)
        logging.debug('response: %s', tlv8.format_string(response_tlv))
        return response_tlv

    return write_http


def perform_pair_setup_part1():
    """
    Performs a pair setup operation as described in chapter 4.7 page 39 ff.

    :return: a tuple of salt and server's public key
    :raises UnavailableError: if the device is already paired
    :raises MaxTriesError: if the device received more than 100 unsuccessful pairing attempts
    :raises BusyError: if a parallel pairing is ongoing
    :raises AuthenticationError: if the verification of the device's SRP proof fails
    :raises MaxPeersError: if the device cannot accept an additional pairing
    :raises IllegalData: if the verification of the accessory's data fails
    """

    #
    # Step #1 ios --> accessory (send SRP start Request) (see page 39)
    #
    logging.debug('#1 ios -> accessory: send SRP start request')
    request_tlv = [
        tlv8.Entry(TlvTypes.State, States.M1),
        tlv8.Entry(TlvTypes.Method, Methods.PairSetup)
    ]

    step2_expectations = {
        TlvTypes.State: tlv8.DataType.INTEGER,
        TlvTypes.Error: tlv8.DataType.INTEGER,
        TlvTypes.PublicKey: tlv8.DataType.BYTES,
        TlvTypes.Salt: tlv8.DataType.BYTES
    }
    response_tlv = yield (request_tlv, step2_expectations)

    #
    # Step #3 ios --> accessory (send SRP verify request) (see page 41)
    #
    logging.debug('#3 ios -> accessory: send SRP verify request')
    assert response_tlv.first_by_id(TlvTypes.State) and response_tlv.first_by_id(
        TlvTypes.State).data == States.M2, 'perform_pair_setup: State not M2'

    # the errors here can be:
    #  * kTLVError_Unavailable: Device is paired
    #  * kTLVError_MaxTries: More than 100 unsuccessful attempts
    #  * kTLVError_Busy: There is already a pairing going on
    error = response_tlv.first_by_id(TlvTypes.Error)
    if error:
        error_handler(error.data, 'step 3 - a')

    assert response_tlv.first_by_id(TlvTypes.PublicKey), 'perform_pair_setup: Not a public key'
    assert response_tlv.first_by_id(TlvTypes.Salt), 'perform_pair_setup: Not a salt'

    return response_tlv.first_by_id(TlvTypes.Salt).data, \
        response_tlv.first_by_id(TlvTypes.PublicKey).data


def perform_pair_setup_part2(pin, ios_pairing_id, salt, server_public_key):
    """
    Performs a pair setup operation as described in chapter 4.7 page 39 ff.

    :param pin: the setup code from the accessory
    :param ios_pairing_id: the id of the simulated ios device
    :return: a dict with the ios device's part of the pairing information
    :raises UnavailableError: if the device is already paired
    :raises MaxTriesError: if the device received more than 100 unsuccessful pairing attempts
    :raises BusyError: if a parallel pairing is ongoing
    :raises AuthenticationError: if the verification of the device's SRP proof fails
    :raises MaxPeersError: if the device cannot accept an additional pairing
    :raises IllegalData: if the verification of the accessory's data fails
    """

    srp_client = SrpClient('Pair-Setup', pin)
    srp_client.set_salt(salt)
    srp_client.set_server_public_key(server_public_key)
    client_pub_key = srp_client.get_public_key()
    client_proof = srp_client.get_proof()

    response_tlv = [
        tlv8.Entry(TlvTypes.State, States.M3),
        tlv8.Entry(TlvTypes.PublicKey, SrpClient.to_byte_array(client_pub_key)),
        tlv8.Entry(TlvTypes.Proof, SrpClient.to_byte_array(client_proof)),
    ]

    step4_expectations = {
        TlvTypes.State: tlv8.DataType.INTEGER,
        TlvTypes.Error: tlv8.DataType.INTEGER,
        TlvTypes.Proof: tlv8.DataType.BYTES
    }
    response_tlv = yield (response_tlv, step4_expectations)

    #
    # Step #5 ios --> accessory (Exchange Request) (see page 43)
    #
    logging.debug('#5 ios -> accessory: send SRP exchange request')

    # M4 Verification (page 43)
    assert response_tlv.first_by_id(TlvTypes.State) and response_tlv.first_by_id(
        TlvTypes.State).data == States.M4, 'perform_pair_setup: State not M4'
    error = response_tlv.first_by_id(TlvTypes.Error)
    if error:
        error_handler(error.data, 'step 5')

    assert response_tlv.first_by_id(TlvTypes.Proof), 'perform_pair_setup: Not a proof'
    if not srp_client.verify_servers_proof(response_tlv.first_by_id(TlvTypes.Proof).data):
        raise AuthenticationError('Step #5: wrong proof!')

    # M5 Request generation (page 44)
    session_key = srp_client.get_session_key()

    ios_device_ltsk = SigningKey.generate()
    ios_device_ltpk = ios_device_ltsk.verify_key

    # reversed:
    #   Pair-Setup-Encrypt-Salt instead of Pair-Setup-Controller-Sign-Salt
    #   Pair-Setup-Encrypt-Info instead of Pair-Setup-Controller-Sign-Info
    hkdf_inst = hkdf.Hkdf('Pair-Setup-Controller-Sign-Salt'.encode(), SrpClient.to_byte_array(session_key),
                          hash=hashlib.sha512)
    ios_device_x = hkdf_inst.expand('Pair-Setup-Controller-Sign-Info'.encode(), 32)

    hkdf_inst = hkdf.Hkdf('Pair-Setup-Encrypt-Salt'.encode(), SrpClient.to_byte_array(session_key),
                          hash=hashlib.sha512)
    session_key = hkdf_inst.expand('Pair-Setup-Encrypt-Info'.encode(), 32)

    ios_device_pairing_id = ios_pairing_id.encode()

    ios_device_info = ios_device_x + ios_device_pairing_id + ios_device_ltpk.encode(encoder=RawEncoder)
    ios_device_signature = ios_device_ltsk.sign(ios_device_info).signature


    sub_tlv = [
        tlv8.Entry(TlvTypes.Identifier, ios_device_pairing_id),
        tlv8.Entry(TlvTypes.PublicKey, ios_device_ltpk.encode(encoder=RawEncoder)),
        tlv8.Entry(TlvTypes.Signature, ios_device_signature)
    ]
    sub_tlv_b = tlv8.encode(sub_tlv)

    # taking tge iOSDeviceX as key was reversed from
    # https://github.com/KhaosT/HAP-NodeJS/blob/2ea9d761d9bd7593dd1949fec621ab085af5e567/lib/HAPServer.js
    # function handlePairStepFive calling encryption.encryptAndSeal
    encrypted_data_with_auth_tag = chacha20_aead_encrypt(bytes(), session_key, 'PS-Msg05'.encode(), bytes([0, 0, 0, 0]),
                                                         sub_tlv_b)
    tmp = bytearray(encrypted_data_with_auth_tag[0])
    tmp += encrypted_data_with_auth_tag[1]

    response_tlv = [
        tlv8.Entry(TlvTypes.State, States.M5),
        tlv8.Entry(TlvTypes.EncryptedData, tmp)
    ]

    step6_expectations = {
        TlvTypes.State: tlv8.DataType.INTEGER,
        TlvTypes.Error: tlv8.DataType.INTEGER,
        TlvTypes.EncryptedData: tlv8.DataType.BYTES
    }
    response_tlv = yield (response_tlv, step6_expectations)

    #
    # Step #7 ios (Verification) (page 47)
    #
    assert response_tlv.first_by_id(TlvTypes.State) and response_tlv.first_by_id(
        TlvTypes.State).data == States.M6, 'perform_pair_setup: State not M6'
    error = response_tlv.first_by_id(TlvTypes.Error)
    if error:
        error_handler(error.data, 'step 7')

    assert response_tlv.first_by_id(TlvTypes.EncryptedData), 'perform_pair_setup: No encrypted data'
    decrypted_data = chacha20_aead_decrypt(bytes(), session_key, 'PS-Msg06'.encode(), bytes([0, 0, 0, 0]),
                                           response_tlv.first_by_id(TlvTypes.EncryptedData).data)
    if decrypted_data is False:
        raise homekit.exception.IllegalData('step 7')

    response_tlv = tlv8.decode(decrypted_data)

    assert response_tlv.first_by_id(TlvTypes.Identifier), 'perform_pair_setup: No identifier'
    accessory_pairing_id = response_tlv.first_by_id(TlvTypes.Identifier).data

    assert response_tlv.first_by_id(TlvTypes.PublicKey), 'perform_pair_setup: No public key'
    accessory_ltpk = response_tlv.first_by_id(TlvTypes.PublicKey).data

    assert response_tlv.first_by_id(TlvTypes.Signature), 'perform_pair_setup: No signature'
    accessory_sig = response_tlv.first_by_id(TlvTypes.Signature).data

    hkdf_inst = hkdf.Hkdf('Pair-Setup-Accessory-Sign-Salt'.encode(),
                          SrpClient.to_byte_array(srp_client.get_session_key()),
                          hash=hashlib.sha512)
    accessory_x = hkdf_inst.expand('Pair-Setup-Accessory-Sign-Info'.encode(), 32)

    accessory_info = accessory_x + accessory_pairing_id + accessory_ltpk

    verify_key = VerifyKey(accessory_ltpk, encoder=RawEncoder)
    try:
        verify_key.verify(accessory_info, accessory_sig)
    except Exception:
        raise InvalidSignatureError('step #7')

    return {
        'AccessoryPairingID': accessory_pairing_id.decode(),
        'AccessoryLTPK': hexlify(accessory_ltpk).decode(),
        'iOSPairingId': ios_pairing_id,
        'iOSDeviceLTSK': ios_device_ltsk.encode(encoder=RawEncoder).hex(),
        'iOSDeviceLTPK': ios_device_ltpk.encode(encoder=RawEncoder).hex()
    }


def get_session_keys(pairing_data):
    """
    HomeKit Controller state machine to perform a pair verify operation as described in chapter 4.8 page 47 ff.
    :param pairing_data: the paring data as returned by perform_pair_setup
    :return: tuple of the session keys (controller_to_accessory_key and  accessory_to_controller_key)
    :raises InvalidAuthTagError: if the auth tag could not be verified,
    :raises IncorrectPairingIdError: if the accessory's LTPK could not be found
    :raises InvalidSignatureError: if the accessory's signature could not be verified
    :raises AuthenticationError: if the secured session could not be established
    """

    #
    # Step #1 ios --> accessory (send verify start Request) (page 47)
    #
    ios_key = x25519.X25519PrivateKey.generate()
    ios_key_pub = ios_key.public_key().public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw
    )

    request_tlv = [
        tlv8.Entry(TlvTypes.State, States.M1),
        tlv8.Entry(TlvTypes.PublicKey, ios_key_pub)
    ]

    step2_expectations = {
        TlvTypes.State: tlv8.DataType.INTEGER,
        TlvTypes.PublicKey: tlv8.DataType.BYTES,
        TlvTypes.EncryptedData: tlv8.DataType.BYTES
    }
    response_tlv = yield (request_tlv, step2_expectations)

    #
    # Step #3 ios --> accessory (send SRP verify request)  (page 49)
    #
    state = response_tlv.first_by_id(TlvTypes.State).data
    assert state == States.M2, 'get_session_keys: not M2'
    assert response_tlv.first_by_id(TlvTypes.PublicKey), 'get_session_keys: no public key'
    assert response_tlv.first_by_id(TlvTypes.EncryptedData), 'get_session_keys: no encrypted data'

    # 1) generate shared secret
    accessorys_session_pub_key_bytes = bytes(response_tlv.first_by_id(TlvTypes.PublicKey).data)
    accessorys_session_pub_key = x25519.X25519PublicKey.from_public_bytes(
        accessorys_session_pub_key_bytes
    )
    shared_secret = ios_key.exchange(accessorys_session_pub_key)

    # 2) derive session key
    hkdf_inst = hkdf.Hkdf('Pair-Verify-Encrypt-Salt'.encode(), shared_secret, hash=hashlib.sha512)
    session_key = hkdf_inst.expand('Pair-Verify-Encrypt-Info'.encode(), 32)

    # 3) verify auth tag on encrypted data and 4) decrypt
    encrypted = response_tlv.first_by_id(TlvTypes.EncryptedData).data
    decrypted = chacha20_aead_decrypt(bytes(), session_key, 'PV-Msg02'.encode(), bytes([0, 0, 0, 0]),
                                      encrypted)
    if type(decrypted) == bool and not decrypted:
        raise InvalidAuthTagError('step 3 - b')
    d1 = tlv8.decode(decrypted)
    assert d1.first_by_id(TlvTypes.Identifier), 'get_session_keys: no identifier'
    assert d1.first_by_id(TlvTypes.Signature), 'get_session_keys: no signature'

    # 5) look up pairing by accessory name
    accessory_name = d1.first_by_id(TlvTypes.Identifier).data.decode()

    if pairing_data['AccessoryPairingID'] != accessory_name:
        raise IncorrectPairingIdError('step 3 - c')

    # 6) verify accessory's signature
    accessory_ltpk = VerifyKey(bytes.fromhex(pairing_data['AccessoryLTPK']), encoder=RawEncoder)
    accessory_sig = d1.first_by_id(TlvTypes.Signature).data
    accessory_session_pub_key_bytes = response_tlv.first_by_id(TlvTypes.PublicKey).data

    # Construct the expected signed data (accessory_info)
    accessory_info = accessory_session_pub_key_bytes + accessory_name.encode() + ios_key_pub

    try:
        accessory_ltpk.verify(accessory_info, accessory_sig)
    except BadSignatureError:
        raise InvalidSignatureError('step 3 - d')

    # 7) create iOSDeviceInfo
    ios_device_info = ios_key_pub + pairing_data['iOSPairingId'].encode() + accessorys_session_pub_key_bytes

    # 8) sign iOSDeviceInfo with long term secret key
    ios_device_ltsk_h = pairing_data['iOSDeviceLTSK']
    ios_device_ltpk_h = pairing_data['iOSDeviceLTPK']

    # PyNaCl only uses the 32-byte secret key (not pubkey)
    ios_device_ltsk = SigningKey(bytes.fromhex(ios_device_ltsk_h))
    ios_device_signature = ios_device_ltsk.sign(ios_device_info).signature

    # 9) construct sub tlv
    sub_tlv = tlv8.encode([
        tlv8.Entry(TlvTypes.Identifier, pairing_data['iOSPairingId'].encode()),
        tlv8.Entry(TlvTypes.Signature, ios_device_signature)
    ])

    # 10) encrypt and sign
    encrypted_data_with_auth_tag = chacha20_aead_encrypt(bytes(), session_key, 'PV-Msg03'.encode(), bytes([0, 0, 0, 0]),
                                                         sub_tlv)
    tmp = bytearray(encrypted_data_with_auth_tag[0])
    tmp += encrypted_data_with_auth_tag[1]

    # 11) create tlv
    request_tlv = [
        tlv8.Entry(TlvTypes.State, States.M3),
        tlv8.Entry(TlvTypes.EncryptedData, tmp)
    ]

    step3_expectations = {
        TlvTypes.State: tlv8.DataType.INTEGER,
        TlvTypes.Error: tlv8.DataType.INTEGER
    }
    response_tlv = yield (request_tlv, step3_expectations)

    #
    #   Post Step #4 verification (page 51)
    #
    state = response_tlv.first_by_id(TlvTypes.State).data

    assert state == States.M4, 'get_session_keys: not M4'

    error = response_tlv.first_by_id(TlvTypes.Error)
    if len(response_tlv) == 2 and error:
        error_handler(error.data, 'verification')

    # calculate session keys
    hkdf_inst = hkdf.Hkdf('Control-Salt'.encode(), shared_secret, hash=hashlib.sha512)
    controller_to_accessory_key = hkdf_inst.expand('Control-Write-Encryption-Key'.encode(), 32)

    hkdf_inst = hkdf.Hkdf('Control-Salt'.encode(), shared_secret, hash=hashlib.sha512)
    accessory_to_controller_key = hkdf_inst.expand('Control-Read-Encryption-Key'.encode(), 32)

    return controller_to_accessory_key, accessory_to_controller_key



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import threading
import select
import logging
import time

from homekit.http_impl.response import HttpResponse
from homekit.crypto.chacha20poly1305 import chacha20_aead_encrypt, chacha20_aead_decrypt
from homekit.http_impl import HttpContentTypes
from homekit import exceptions


class SecureHttp:
    """
    Class to help in the handling of HTTP requests and responses that are performed following chapter 5.5 page 70ff of
    the HAP specification.
    """

    def __init__(self, session, timeout=10):
        """
        Initializes the secure HTTP class. The required keys can be obtained with get_session_keys

        :param sock: the socket over which the communication takes place
        :param a2c_key: the key used for the communication between accessory and controller
        :param c2a_key: the key used for the communication between controller and accessory
        """
        self.sock = session.sock
        self.host = session.pairing_data['AccessoryIP']
        self.port = session.pairing_data['AccessoryPort']
        self.a2c_key = session.a2c_key
        self.c2a_key = session.c2a_key
        self.c2a_counter = 0
        self.a2c_counter = 0
        self.timeout = timeout
        self.lock = threading.Lock()

    def get(self, target):
        data = 'GET {tgt} HTTP/1.1\nHost: {host}:{port}\n\n'.format(tgt=target, host=self.host, port=self.port)
        data = data.replace("\n", "\r\n")
        return self._handle_request(data.encode())

    def put(self, target, body, content_type=HttpContentTypes.JSON):
        headers = 'Host: {host}:{port}\n'.format(host=self.host, port=self.port) + \
                  'Content-Type: {ct}\n'.format(ct=content_type) + \
                  'Content-Length: {len}\n'.format(len=len(body))
        data = 'PUT {tgt} HTTP/1.1\n{hdr}\n{body}'.format(tgt=target, hdr=headers, body=body)
        data = data.replace("\n", "\r\n")
        return self._handle_request(data.encode())

    def post(self, target, body, content_type=HttpContentTypes.TLV):
        headers = 'Host: {host}:{port}\n'.format(host=self.host, port=self.port) + \
                  'Content-Type: {ct}\n'.format(ct=content_type) + \
                  'Content-Length: {len}\n'.format(len=len(body))
        data = 'POST {tgt} HTTP/1.1\n{hdr}\n'.format(tgt=target, hdr=headers)
        data = data.replace("\n", "\r\n")
        return self._handle_request(data.encode() + body)

    def _handle_request(self, data):
        logging.debug('handle request: %s', data)
        with self.lock:
            while len(data) > 0:
                # split the data to max 1024 bytes (see page 71)
                len_data = min(len(data), 1024)
                tmp_data = data[:len_data]
                data = data[len_data:]
                len_bytes = len_data.to_bytes(2, byteorder='little')
                cnt_bytes = self.c2a_counter.to_bytes(8, byteorder='little')
                self.c2a_counter += 1
                ciper_and_mac = chacha20_aead_encrypt(len_bytes, self.c2a_key, cnt_bytes, bytes([0, 0, 0, 0]),
                                                      tmp_data)

                try:
                    self.sock.send(len_bytes + ciper_and_mac[0] + ciper_and_mac[1])
                except OSError as e:
                    raise exceptions.AccessoryDisconnectedError(str(e))

            return self._read_response(self.timeout)

    def _read_response(self, timeout=10):
        tmp = bytearray()
        response = HttpResponse()
        self.sock.setblocking(0)
        start_time = time.monotonic()

        while not response.is_read_completely():
            if timeout != 0:
                elapsed = time.monotonic() - start_time
                remaining = max(0, timeout - elapsed)
                if remaining == 0:
                    raise TimeoutError("Timed out waiting for response.")
            else:
                remaining = None  # wait forever

            rlist, _, _ = select.select([self.sock], [], [], remaining)
            if not rlist:
                continue  # No data yet, loop again

            try:
                if len(tmp) < 2:
                    tmp += self.sock.recv(2 - len(tmp))
                    continue

                length = int.from_bytes(tmp[:2], 'little')
                expected_total = 2 + length + 16  # 2-byte length + data + tag
                while len(tmp) < expected_total:
                    tmp += self.sock.recv(expected_total - len(tmp))

                tmp = tmp[2:]
                block = tmp[:length]
                tag = tmp[length:length+16]
                tmp = tmp[length+16:]

                decrypted = self.decrypt_block(length, block, tag)
                if decrypted is not False:
                    response.parse(decrypted)
                else:
                    self.sock.close()
                    raise exceptions.EncryptionError('Error during transmission.')

            except BlockingIOError:
                continue

        return response

    def decrypt_block(self, length, block, tag):
        tmp = chacha20_aead_decrypt(length.to_bytes(2, byteorder='little'),
                                    self.a2c_key,
                                    self.a2c_counter.to_bytes(8, byteorder='little'),
                                    bytes([0, 0, 0, 0]), block + tag)
        if tmp is not False:
            self.a2c_counter += 1

        return tmp

    def handle_event_response(self, timeout: int| None = None):
        """
        This reads the enciphered response from an accessory after registering for events.
        :param timeout: how many seconds to wait for the next event, or None for infinite
        """
        try:
            return self._read_response(self.timeout if timeout is None else timeout)
        except OSError as e:
            raise exceptions.AccessoryDisconnectedError(str(e))



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import sys
import argparse
import logging
import tlv8
import base64

from homekit.controller import Controller
from homekit.model.characteristics import CharacteristicsTypes, CharacteristicsDecoderLoader
from homekit.model.services import ServicesTypes
from homekit.log_support import setup_logging, add_log_arguments
from homekit.model.characteristics.characteristic_formats import CharacteristicFormats


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit get accessories app')
    parser.add_argument('-f', action='store', required=True, dest='file', help='File with the pairing data')
    parser.add_argument('-a', action='store', required=True, dest='alias', help='alias for the pairing')
    parser.add_argument('-o', action='store', dest='output', default='compact', choices=['json', 'compact'],
                        help='Specify output format')
    parser.add_argument('--adapter', action='store', dest='adapter', default='hci0',
                        help='the bluetooth adapter to be used (defaults to hci0)')
    parser.add_argument('-d', action='store_true', required=False, dest='decode',
                        help='If set, try to find a decoder for each characteristic and show the decoded value')
    add_log_arguments(parser)
    return parser.parse_args()


def decode_values(input_data):
    """
    Try to decode the values from the given data structure. This depends on:
     - if the field is of type `CharacteristicFormats.tlv8`
     - a decoder was found for the characteristic

    :param input_data: data returned from `AbstractPairing.list_accessories_and_characteristics`
    :return: the input data with decoded value fields
    """
    loader = CharacteristicsDecoderLoader()
    for accessory in input_data:
        for service in accessory['services']:
            for characteristic in service['characteristics']:
                c_format = characteristic['format']
                c_type = characteristic['type']
                # TODO what about CharacteristicFormats.data?
                if c_format in [CharacteristicFormats.tlv8]:
                    decoder = loader.load(c_type)
                    if decoder:
                        try:
                            characteristic['value'] = decoder(base64.b64decode(characteristic['value']))
                        except Exception as e:
                            logging.error('error during decode %s: %s' % (c_type, characteristic['value']), e)
    return input_data


if __name__ == '__main__':
    args = setup_args_parser()

    setup_logging(args.loglevel)

    controller = Controller(args.adapter)
    try:
        controller.load_data(args.file)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    if args.alias not in controller.get_pairings():
        print('"{a}" is no known alias'.format(a=args.alias))
        sys.exit(-1)

    try:
        pairing = controller.get_pairings()[args.alias]
        data = pairing.list_accessories_and_characteristics()
        controller.save_data(args.file)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    if args.decode:
        data = decode_values(data)

    # prepare output
    if args.output == 'json':
        print(json.dumps(data, indent=4, cls=tlv8.JsonEncoder))

    if args.output == 'compact':
        for accessory in data:
            aid = accessory['aid']
            for service in accessory['services']:
                s_type = service['type']
                s_iid = service['iid']
                print('{aid}.{iid}: >{stype}<'.format(aid=aid, iid=s_iid, stype=ServicesTypes.get_short(s_type)))

                for characteristic in service['characteristics']:
                    c_iid = characteristic['iid']
                    value = characteristic.get('value', '')
                    c_type = characteristic['type']
                    c_format = characteristic['format']
                    # we need to get the entry list from the decoder into a string and reformat it for better
                    # readability.
                    if args.decode and c_format in [CharacteristicFormats.tlv8] and isinstance(value, tlv8.EntryList):
                        value = tlv8.format_string(value)
                        value = '      '.join(value.splitlines(keepends=True))
                        value = '\n      ' + value
                    perms = ','.join(characteristic['perms'])
                    desc = characteristic.get('description', '')
                    c_type = CharacteristicsTypes.get_short(c_type)
                    print('  {aid}.{iid}: ({description}) >{ctype}< [{perms}]'.format(aid=aid,
                                                                                      iid=c_iid,
                                                                                      ctype=c_type,
                                                                                      perms=perms,
                                                                                      description=desc))
                    print('    Value: {value}'.format(value=value))



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class HomeKitException(Exception):
    """Generic HomeKit exception.
    Attributes:
        stage: the stage that the exception occurred at
    """

    def __init__(self, stage):
        self.stage = stage

    pass


class BluetoothAdapterError(HomeKitException):
    pass


class MalformedPinError(HomeKitException):
    """
    Class to represent a malformed pin (not following the form DDD-DD-DDD)
    """
    pass


class ProtocolError(HomeKitException):
    """
    Class to represent an abstraction layer for all errors that are defined in the Error Codes table 4-5 page 60 of the
    specification.
    """
    pass


class UnknownError(ProtocolError):
    """
    Raised upon receipt of an unknown error (transmission of Errors.Unknown). The spec says that this can happen
    during "Add Pairing" (chapter 4.11 page 51) and "Remove Pairing" (chapter 4.12 page 53).
    """
    pass


class AuthenticationError(ProtocolError):
    """
    Raised upon receipt of an authentication error. This can happen on:
     * multiple occasions through out setup pairing (M4 / page 42, M5 page 45): if the pairing could not be established
     * during pair verify (M4 / page 50): if the session key could not be generated
     * during add pair (M2 / page 52): if the controller is not admin
     * during remove pairing (M2 / page 54): if the controller is not admin
     * during list pairing (M2 / page 56): if the controller is not admin
    """
    pass


class BackoffError(ProtocolError):
    """
    Raised upon receipt of a back off error. It seems unclear when this is raised, must be related to
    TlvType.RetryDelay which is defined on page 61 of the spec.
    """
    pass


class MaxPeersError(ProtocolError):
    """
    Raised upon receipt of a max peers error. This can happen:
     * during executing a "pair setup" command
     * during an "add pairing" command
    """
    pass


class MaxTriesError(ProtocolError):
    """
    Raised upon receipt of a max tries error during a pair setup procedure. This happens if more than 100 unsuccessful
    authentication attempts were performed.
    """
    pass


class UnavailableError(ProtocolError):
    """Raised upon receipt of an unavailable error"""
    pass


class BusyError(ProtocolError):
    """
    Raised upon receipt of a busy error during a pair setup procedure. This happens only if a parallel pairing process
    is ongoing.
    """
    pass


class InvalidError(ProtocolError):
    """
    Raised upon receipt of an error not defined in the HomeKit spec. This should basically never be raised since it is
    the default error in the protocol's error handler.
    """
    pass


class HttpException(Exception):
    """
    Used within the HTTP Parser.
    """
    def __init__(self, message):
        Exception.__init__(self, message)


class InvalidAuthTagError(ProtocolError):
    """
    Raised upon receipt of an invalid auth tag in Pair Verify Step 3.3 (Page 49).
    """
    pass


class IncorrectPairingIdError(ProtocolError):
    """
    Raised in Pair Verify Step 3.5 (Page 49) if the accessory responds with an unexpected pairing id.
    """
    pass


class InvalidSignatureError(ProtocolError):
    """
    Raised upon receipt of an invalid signature either from an accessory or from the controller.
    """
    pass


class ConfigurationError(HomeKitException):
    """
    Used if any configuration in the HomeKit AccessoryServer's context was wrong.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class FormatError(HomeKitException):
    """
    Used if any format conversion fails or is impossible.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class CharacteristicPermissionError(HomeKitException):
    """
    Used if the characteristic's permissions do not allow the action. This includes reads on write only characteristics
    and writes on read only characteristics.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class AccessoryNotFoundError(HomeKitException):
    """
    Used if a HomeKit Accessory's IP and port could not be received via Bonjour / Zeroconf. This might be a temporary
    issue due to the way Bonjour / Zeroconf works.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class EncryptionError(HomeKitException):
    """
    Used if during a transmission some errors occurred.
    """
    def __init__(self, message):
        Exception.__init__(self, message)


class AccessoryDisconnectedError(HomeKitException):
    """
    Used if a HomeKit disconnects part way through an operation or series of operations.

    It may be possible to reconnect and retry the request.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class ConfigLoadingError(HomeKitException):
    """
    Used on problems loading some config. This includes but may not be limited to:
     * problems with file permissions (file not readable)
     * the file could not be found
     * the file does not contain parseable JSON
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class ConfigSavingError(HomeKitException):
    """
    Used on problems saving some config. This includes but may not be limited to:
     * problems with file permissions (file not writable)
     * the file could not be found (occurs if the path does not exist)
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class UnpairedError(HomeKitException):
    """
    This should be raised if a paired accessory is expected but the accessory is still unpaired.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class AlreadyPairedError(HomeKitException):
    """
    This should be raised if an unpaired accessory is expected but the accessory is already paired.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class RequestRejected(HomeKitException):
    """
    Raised when a request fails with a HAP error code
    """

    def __init__(self, message, error_code):
        self.error_code = error_code
        self.message = message
        Exception.__init__(message)


class TransportNotSupportedError(HomeKitException):
    def __init__(self, transport):
        Exception.__init__(self,
                           'Transport {t} not supported. See setup.py for required dependencies.'.format(t=transport))


class DisconnectedControllerError(HomeKitException):
    def __init__(self):
        Exception.__init__(self, 'Controller has passed away')



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import abc
import base64
import binascii
import tlv8

from homekit.exceptions import FormatError
from homekit.model.characteristics import CharacteristicFormats



def strtobool(val):
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    if val in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    raise ValueError(f"invalid truth value {val!r}")


class AbstractPairing(abc.ABC):

    def _get_pairing_data(self):
        """
        This method returns the internal pairing data. DO NOT mess around with it.

        :return: a dict containing the data
        """
        return self.pairing_data

    @abc.abstractmethod
    def close(self):
        """
        Close the pairing's communications. This closes the session.
        """
        pass

    @abc.abstractmethod
    def list_accessories_and_characteristics(self):
        """
        This retrieves a current set of accessories and characteristics behind this pairing.

        :return: the accessory data as described in the spec on page 73 and following
        :raises AccessoryNotFoundError: if the device can not be found via zeroconf
        """
        pass

    @abc.abstractmethod
    def list_pairings(self):
        """
        This method returns all pairings of a HomeKit accessory. This always includes the local controller and can only
        be done by an admin controller.

        The keys in the resulting dicts are:
         * pairingId: the pairing id of the controller
         * publicKey: the ED25519 long-term public key of the controller
         * permissions: bit value for the permissions
         * controllerType: either admin or regular

        :return: a list of dicts sorted by the elements pairingIds
        :raises: UnknownError: if it receives unexpected data
        :raises: UnpairedError: if the polled accessory is not paired
        """
        pass

    @abc.abstractmethod
    def get_characteristics(self, characteristics, include_meta=False, include_perms=False, include_type=False,
                            include_events=False):
        """
        This method is used to get the current readouts of any characteristic of the accessory.

        :param characteristics: a list of 2-tupels of accessory id and instance id
        :param include_meta: if True, include meta information about the characteristics. This contains the format and
                             the various constraints like maxLen and so on.
        :param include_perms: if True, include the permissions for the requested characteristics.
        :param include_type: if True, include the type of the characteristics in the result. See CharacteristicsTypes
                             for translations.
        :param include_events: if True on a characteristics that supports events, the result will contain information if
                               the controller currently is receiving events for that characteristic. Key is 'ev'.
        :return: a dict mapping 2-tupels of aid and iid to dicts with value or status and description, e.g.
                 {(1, 8): {'value': 23.42}
                  (1, 37): {'description': 'Resource does not exist.', 'status': -70409}
                 }
        """
        pass

    @abc.abstractmethod
    def put_characteristics(self, characteristics, do_conversion=False):
        """
        Update the values of writable characteristics. The characteristics have to be identified by accessory id (aid),
        instance id (iid). If do_conversion is False (the default), the value must be of proper format for the
        characteristic since no conversion is done. If do_conversion is True, the value is converted.

        :param characteristics: a list of 3-tupels of accessory id, instance id and the value
        :param do_conversion: select if conversion is done (False is default)
        :return: a dict from (aid, iid) onto {status, description}
        :raises FormatError: if the input value could not be converted to the target type and conversion was
                             requested
        """
        pass

    @abc.abstractmethod
    def get_events(self, characteristics, callback_fun, max_events=-1, max_seconds=-1):
        """
        This function is called to register for events on characteristics and receive them. Each time events are
        received a call back function is invoked. By that the caller gets information about the events.

        The characteristics are identified via their proper accessory id (aid) and instance id (iid).

        The call back function takes a list of 3-tupels of aid, iid and the value, e.g.:
          [(1, 9, 26.1), (1, 10, 30.5)]

        If the input contains characteristics without the event permission or any other error, the function will return
        a dict containing tupels of aid and iid for each requested characteristic with error. Those who would have
        worked are not in the result.

        :param characteristics: a list of 2-tupels of accessory id (aid) and instance id (iid)
        :param callback_fun: a function that is called each time events were recieved
        :param max_events: number of reported events, default value -1 means unlimited
        :param max_seconds: number of seconds to wait for events, default value -1 means unlimited
        :return: a dict mapping 2-tupels of aid and iid to dicts with status and description, e.g.
                 {(1, 37): {'description': 'Notification is not supported for characteristic.', 'status': -70406}}
        """
        pass

    @abc.abstractmethod
    def identify(self):
        """
        This call can be used to trigger the identification of a paired accessory. A successful call should
        cause the accessory to perform some specific action by which it can be distinguished from the others (blink a
        LED for example).

        It uses the identify characteristic as described on page 152 of the spec.

        :return True, if the identification was run, False otherwise
        """
        pass

    @abc.abstractmethod
    def add_pairing(self, additional_controller_pairing_identifier, ios_device_ltpk, permissions):
        pass


class NotSupportedPairing(AbstractPairing):
    """
    Implementation of `AbstractPairing` with no implemented functions but only for handling connection type and pairing
    data. This is used in `Controller.load_data` if the connection type of the pairing is not supported (e.g. exchanging
    a file with pairing data between linux (IP and BLE) and OS X (IP only). Without using this `DummyPairing` the
    pairings would be lost on saving of the file.
    """

    def __init__(self, pairing_data, connection_type):
        self.pairing_data = pairing_data
        self.connection_type = connection_type

    def close(self):
        pass

    def list_accessories_and_characteristics(self):
        raise NotImplementedError(
            'Connection type "{}" is not supported in this setup!'.format(self.connection_type))

    def list_pairings(self):
        raise NotImplementedError(
            'Connection type "{}" is not supported in this setup!'.format(self.connection_type))

    def get_characteristics(self, characteristics, include_meta=False, include_perms=False, include_type=False,
                            include_events=False):
        raise NotImplementedError(
            'Connection type "{}" is not supported in this setup!'.format(self.connection_type))

    def put_characteristics(self, characteristics, do_conversion=False):
        raise NotImplementedError(
            'Connection type "{}" is not supported in this setup!'.format(self.connection_type))

    def get_events(self, characteristics, callback_fun, max_events=-1, max_seconds=-1):
        raise NotImplementedError(
            'Connection type "{}" is not supported in this setup!'.format(self.connection_type))

    def identify(self):
        raise NotImplementedError(
            'Connection type "{}" is not supported in this setup!'.format(self.connection_type))

    def add_pairing(self, additional_controller_pairing_identifier, ios_device_ltpk, permissions):
        raise NotImplementedError(
            'Connection type "{}" is not supported in this setup!'.format(self.connection_type))


def check_convert_value(val, target_type):
    """
    Checks if the given value is of the given type or is convertible into the type. If the value is not convertible, a
    HomeKitTypeException is thrown.

    :param val: the original value
    :param target_type: the target type of the conversion
    :return: the converted value
    :raises FormatError: if the input value could not be converted to the target type
    """
    if target_type == CharacteristicFormats.bool:
        try:
            val = strtobool(str(val))
        except ValueError:
            raise FormatError('"{v}" is no valid "{t}"!'.format(v=val, t=target_type))
    if target_type in [CharacteristicFormats.uint64, CharacteristicFormats.uint32,
                       CharacteristicFormats.uint16, CharacteristicFormats.uint8,
                       CharacteristicFormats.int]:
        try:
            val = int(val)
        except ValueError:
            raise FormatError('"{v}" is no valid "{t}"!'.format(v=val, t=target_type))
    if target_type == CharacteristicFormats.float:
        try:
            val = float(val)
        except ValueError:
            raise FormatError('"{v}" is no valid "{t}"!'.format(v=val, t=target_type))
    if target_type == CharacteristicFormats.data:
        try:
            base64.decodebytes(val.encode())
        except binascii.Error:
            raise FormatError('"{v}" is no valid "{t}"!'.format(v=val, t=target_type))
    if target_type == CharacteristicFormats.tlv8:
        try:
            tmp_bytes = base64.decodebytes(val.encode())
            tlv8.decode(tmp_bytes, {})
        except (binascii.Error, ValueError):
            raise FormatError('"{v}" is no valid "{t}"!'.format(v=val, t=target_type))
    return val



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Utility to add an additional paring to a HomeKit Accessory (see chapter 4.11 page 51ff)
"""

import argparse
import sys
import logging

from homekit.controller import Controller
from homekit.log_support import setup_logging, add_log_arguments
from homekit.exceptions import HomeKitException


def setup_args_parser():
    """
    Setup the parser for the CLI parameters and perform the parsing.

    :return: an instance of argparse.Namespace containing the parsed parameters
    """
    parser = argparse.ArgumentParser(description='HomeKit generate pairing data app')
    parser.add_argument('-f', action='store', required=True, dest='file', help='HomeKit pairing data file')
    parser.add_argument('-a', action='store', required=True, dest='alias', help='alias for the pairing')
    parser.add_argument('-i', action='store', required=True, dest='pairing_id',
                        help='the device identification of the other controller')
    parser.add_argument('-k', action='store', required=True, dest='key',
                        help='the public key of the other controller')
    parser.add_argument('-p', action='store', required=True, dest='permission', choices=['User', 'Admin'],
                        help='the privilege of the other controller')
    parser.add_argument('--adapter', action='store', dest='adapter', default='hci0',
                        help='the bluetooth adapter to be used (defaults to hci0)')
    add_log_arguments(parser)
    return parser.parse_args()


def main():
    """
    the main function.

    :return: none
    """
    args = setup_args_parser()

    setup_logging(args.loglevel)

    controller = Controller()
    try:
        controller.load_data(args.file)
    except HomeKitException as exception:
        print(exception)
        logging.debug(exception, exc_info=True)
        sys.exit(-1)

    if args.alias not in controller.get_pairings():
        print('"{a}" is no known alias'.format(a=args.alias))
        sys.exit(-1)

    try:
        pairing = controller.get_pairings()[args.alias]
        pairing.add_pairing(args.pairing_id, args.key, args.permission)
        print(pairing.pairing_data)

        if pairing.pairing_data['Connection'] == 'IP':
            text = 'Please add this to homekit.finish_add_remote_pairing:\n' \
                   '    -c {c} -i {id} -k {pk}' \
                .format(c=pairing.pairing_data['Connection'],
                        id=pairing.pairing_data['AccessoryPairingID'],
                        pk=pairing.pairing_data['AccessoryLTPK']
                        )
            print(text)

        else:
            print('Unknown pairing data connection type')
    except HomeKitException as exception:
        print(exception)
        logging.debug(exception, exc_info=True)
        sys.exit(-1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import sys
import logging
import uuid
from nacl.signing import SigningKey
from binascii import hexlify

from homekit.controller import Controller
from homekit.controller.additional_pairing import AdditionalPairing
from homekit.log_support import setup_logging, add_log_arguments
from homekit.exceptions import AlreadyPairedError


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit generate pairing data app')
    parser.add_argument('-f', action='store', required=True, dest='file', help='HomeKit pairing data file')
    parser.add_argument('-a', action='store', required=True, dest='alias', help='alias for the pairing')
    add_log_arguments(parser)
    return parser.parse_args()


if __name__ == '__main__':
    args = setup_args_parser()

    setup_logging(args.loglevel)

    controller = Controller()
    try:
        controller.load_data(args.file)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    try:
        pairings = controller.get_pairings()
        if args.alias in pairings:
            pairing_data = pairings[args.alias]._get_pairing_data()
            additional_controller_pairing_identifier = pairing_data['iOSPairingId']
            ios_device_ltpk = pairing_data['iOSDeviceLTPK']
            text = 'Alias "{a}" is already in state add additional pairing.\n'\
                   'Please add this to homekit.add_additional_pairing:\n'\
                   '    -i {id} -k {pk}'\
                   .format(a=args.alias,
                           id=additional_controller_pairing_identifier,
                           pk=ios_device_ltpk
                           )
            raise AlreadyPairedError(text)

        additional_controller_pairing_identifier = str(uuid.uuid4())
        ios_device_ltsk = SigningKey.generate()
        ios_device_ltpk = ios_device_ltsk.verify_key

        # Format the keys
        public_key = hexlify(ios_device_ltpk.encode()).decode()
        private_key = hexlify(ios_device_ltsk.encode()).decode()

        text = 'Please add this to homekit.add_additional_pairing:\n' \
               '    -i {id} -k {pk}' \
            .format(id=additional_controller_pairing_identifier,
                    pk=public_key
                    )
        print(text)

        a = {
            'iOSPairingId': additional_controller_pairing_identifier,
                'iOSDeviceLTSK': private_key[:64],  # 32-byte hex-encoded private key
                'iOSDeviceLTPK': public_key,        # 32-byte hex-encoded public key
            'Connection': 'ADDITIONAL_PAIRING'
        }
        pairings[args.alias] = AdditionalPairing(a)
        controller.save_data(args.file)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import locale

from homekit.log_support import setup_logging, add_log_arguments
from homekit.controller import Controller


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit IP discover app -'
                                                 ' list all HomeKit devices on the same IP network')
    parser.add_argument('-t', action='store', required=False, dest='timeout', type=int, default=10,
                        help='Number of seconds to wait')
    parser.add_argument('-u', action='store_true', required=False, dest='unpaired_only',
                        help='If activated, this option will show only unpaired HomeKit IP Devices')
    add_log_arguments(parser)
    return parser.parse_args()


def prepare_string(input_string):
    """
    Make a string save for printing in a terminal. The string get recoded using the terminals preferred locale and
    replacing the characters that cannot be encoded.
    :param input_string: the input string
    :return: the output string which is save for printing
    """
    return '{t}'.format(t=input_string.encode(locale.getpreferredencoding(), errors='replace').decode())


if __name__ == '__main__':
    args = setup_args_parser()
    setup_logging(args.loglevel)

    results = Controller.discover(args.timeout)
    for info in results:
        if args.unpaired_only and info['sf'] == '0':
            continue
        print('Name: {name}'.format(name=prepare_string(info['name'])))
        print('Url: http_impl://{ip}:{port}'.format(ip=info['address'], port=info['port']))
        print('Configuration number (c#): {conf}'.format(conf=info['c#']))
        print('Feature Flags (ff): {f} (Flag: {flags})'.format(f=info['flags'], flags=info['ff']))
        print('Device ID (id): {id}'.format(id=info['id']))
        print('Model Name (md): {md}'.format(md=prepare_string(info['md'])))
        print('Protocol Version (pv): {pv}'.format(pv=info['pv']))
        print('State Number (s#): {sn}'.format(sn=info['s#']))
        print('Status Flags (sf): {sf} (Flag: {flags})'.format(sf=info['statusflags'], flags=info['sf']))
        print('Category Identifier (ci): {c} (Id: {ci})'.format(c=info['category'], ci=info['ci']))
        print()



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json

id_counter = 0


def get_id():
    global id_counter
    id_counter += 1
    return id_counter


class ToDictMixin(object):
    """
    Will help to convert the various accessories, services and characteristics to JSON.
    """

    def _to_dict(self):
        tmp = {}
        for x in dir(self):
            if x.startswith('_') or callable(getattr(self, x)):
                continue
            val = getattr(self, x)
            if val is None:
                continue
            if isinstance(val, list):
                tmpval = []
                for e in val:
                    if isinstance(e, str):
                        tmpval.append(e)
                    else:
                        tmpval.append(e._to_dict())
                tmp[x] = tmpval
            else:
                tmp[x] = val

        return tmp

    def __str__(self):
        d = self._to_dict()
        return json.dumps(d)



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    'ThermostatService', 'LightBulbService', 'FanService', 'BHSLightBulbService', 'AccessoryInformationService',
    'OutletService', 'AbstractService', 'ServicesTypes'
]

from homekit.model.services.service_types import ServicesTypes

from homekit.model.services.abstract_service import AbstractService
from homekit.model.services.accessoryinformation_service import AccessoryInformationService
from homekit.model.services.bhslightbulb_service import BHSLightBulbService
from homekit.model.services.fan_service import FanService
from homekit.model.services.lightbulb_service import LightBulbService
from homekit.model.services.outlet_service import OutletService
from homekit.model.services.thermostat_service import ThermostatService



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
from json.decoder import JSONDecodeError
import logging
import random
import uuid
import re
import tlv8

from homekit.exceptions import AccessoryNotFoundError, ConfigLoadingError, UnknownError, \
    AuthenticationError, ConfigSavingError, AlreadyPairedError, TransportNotSupportedError, MalformedPinError
from homekit.protocol import States, Methods, Errors, TlvTypes
from homekit.http_impl import HomeKitHTTPConnection
from homekit.protocol.statuscodes import HapStatusCodes
from homekit.protocol import perform_pair_setup_part1, perform_pair_setup_part2, create_ip_pair_setup_write
from homekit.model.services.service_types import ServicesTypes
from homekit.model.characteristics.characteristic_types import CharacteristicsTypes
from homekit.protocol.opcodes import HapBleOpCodes
from homekit.controller.tools import NotSupportedPairing
from homekit.controller.additional_pairing import AdditionalPairing

from homekit.zeroconf_impl import discover_homekit_devices, find_device_ip_and_port
from homekit.controller.ip_implementation import IpPairing, IpSession


class Controller(object):
    """
    This class represents a HomeKit controller (normally your iPhone or iPad).
    """

    def __init__(self):
        """
        Initialize an empty controller. Use 'load_data()' to load the pairing data.

        """
        self.pairings = {}
        self.logger = logging.getLogger('homekit.controller.Controller')

    @staticmethod
    def discover(max_seconds=10):
        """
        Perform a Bonjour discovery for HomeKit accessory. The discovery will last for the given amount of seconds. The
        result will be a list of dicts. The keys of the dicts are:
         * name: the Bonjour name of the HomeKit accessory (i.e. Testsensor1._hap._tcp.local.)
         * address: the IP address of the accessory
         * port: the used port
         * c#: the configuration number (required)
         * ff / flags: the numerical and human readable version of the feature flags (supports pairing or not, see table
                       5-8 page 69)
         * id: the accessory's pairing id (required)
         * md: the model name of the accessory (required)
         * pv: the protocol version
         * s#: the current state number (required)
         * sf / statusflags: the status flag (see table 5-9 page 70)
         * ci / category: the category identifier in numerical and human readable form. For more information see table
                        12-3 page 254 or homekit.Categories (required)

        IMPORTANT:
        This method will ignore all HomeKit accessories that exist in _hap._tcp domain but fail to have all required
        TXT record keys set.

        :param max_seconds: how long should the Bonjour service browser do the discovery (default 10s). See sleep for
                            more details
        :return: a list of dicts as described above
        """
        return discover_homekit_devices(max_seconds)

    @staticmethod
    def identify(accessory_id):
        """
        This call can be used to trigger the identification of an accessory, that was not yet paired. A successful call
        should cause the accessory to perform some specific action by which it can be distinguished from others (blink a
        LED for example).

        It uses the /identify url as described on page 88 of the spec.

        :param accessory_id: the accessory's pairing id (e.g. retrieved via discover)
        :raises AccessoryNotFoundError: if the accessory could not be looked up via Bonjour
        :raises AlreadyPairedError: if the accessory is already paired
        """
        connection_data = find_device_ip_and_port(accessory_id)
        if connection_data is None:
            raise AccessoryNotFoundError('Cannot find accessory with id "{i}".'.format(i=accessory_id))

        conn = HomeKitHTTPConnection(connection_data['ip'], port=connection_data['port'])
        conn.request('POST', '/identify')
        resp = conn.getresponse()

        # spec says status code 400 on any error (page 88). It also says status should be -70401 (which is "Request
        # denied due to insufficient privileges." table 5-12 page 80) but this sounds odd.
        if resp.code == 400:
            data = json.loads(resp.read().decode())
            code = data['status']
            conn.close()
            raise AlreadyPairedError(
                'identify failed because: {reason} ({code}).'.format(reason=HapStatusCodes[code],
                                                                     code=code))
        conn.close()

    def shutdown(self):
        """
        Shuts down the controller by closing all connections that might be held open by the pairings of the controller.
        """
        for p in self.pairings:
            self.pairings[p].close()

    def get_pairings(self):
        """
        Returns a dict containing all pairings known to the controller.

        :return: the dict maps the aliases to Pairing objects
        """
        return self.pairings

    def load_data(self, filename):
        """
        Loads the pairing data of the controller from a file. If the connection type of a pairing is currently not
        supported by the python environment (e.g. because of missing modules for BLE), an instance of
        `NotSupportedPairing` is used. By this no pairings are lost during a `load_data`/`save_data` cycle.

        :param filename: the file name of the pairing data
        :raises ConfigLoadingError: if the config could not be loaded. The reason is given in the message.
        :raises TransportNotSupportedError: if the dependencies for the selected transport are not installed
        """
        try:
            with open(filename, 'r') as input_fp:
                data = json.load(input_fp)
                for pairing_id in data:
                    if 'Connection' not in data[pairing_id]:
                        # This is a pre BLE entry in the file with the pairing data, hence it is for an IP based
                        # accessory. So we set the connection type (in case save data is used everything will be fine)
                        # and also issue a warning
                        data[pairing_id]['Connection'] = 'IP'
                        self.logger.warning(
                            'Loaded pairing for %s with missing connection type. Assume this is IP based.', pairing_id)

                    if data[pairing_id]['Connection'] == 'IP':
                        self.pairings[pairing_id] = IpPairing(data[pairing_id])
                    elif data[pairing_id]['Connection'] == 'ADDITIONAL_PAIRING':
                        self.pairings[pairing_id] = AdditionalPairing(data[pairing_id])
                    else:
                        # ignore anything else, issue warning
                        self.logger.warning('could not load pairing %s of type "%s"', pairing_id,
                                            data[pairing_id]['Connection'])
        except PermissionError:
            raise ConfigLoadingError('Could not open "{f}" due to missing permissions'.format(f=filename))
        except JSONDecodeError:
            raise ConfigLoadingError('Cannot parse "{f}" as JSON file'.format(f=filename))
        except FileNotFoundError:
            raise ConfigLoadingError('Could not open "{f}" because it does not exist'.format(f=filename))

    def save_data(self, filename):
        """
        Saves the pairing data of the controller to a file.

        :param filename: the file name of the pairing data
        :raises ConfigSavingError: if the config could not be saved. The reason is given in the message.
        """
        data = {}
        for pairing_id in self.pairings:
            # package visibility like in java would be nice here
            data[pairing_id] = self.pairings[pairing_id]._get_pairing_data()
        try:
            with open(filename, 'w') as output_fp:
                json.dump(data, output_fp, indent='  ')
        except PermissionError:
            raise ConfigSavingError('Could not write "{f}" due to missing permissions'.format(f=filename))
        except FileNotFoundError:
            raise ConfigSavingError(
                'Could not write "{f}" because it (or the folder) does not exist'.format(f=filename))

    @staticmethod
    def check_pin_format(pin):
        """
        Checks the format of the given pin: XXX-XX-XXX with X being a digit from 0 to 9

        :raises MalformedPinError: if the validation fails
        """
        if not re.match(r'^\d\d\d-\d\d-\d\d\d$', pin):
            raise MalformedPinError('The pin must be of the following XXX-XX-XXX where X is a digit between 0 and 9.')

    def perform_pairing(self, alias, accessory_id, pin):
        """
        This performs a pairing attempt with the IP accessory identified by its id.

        Accessories can be found via the discover method. The id field is the accessory's id for the second parameter.

        The required pin is either printed on the accessory or displayed. Must be a string of the form 'XXX-YY-ZZZ'.

        Important: no automatic saving of the pairing data is performed. If you don't do this, the information is lost
            and you have to reset the accessory!

        :param alias: the alias for the accessory in the controllers data
        :param accessory_id: the accessory's id
        :param pin: function to return the accessory's pin
        :raises AccessoryNotFoundError: if no accessory with the given id can be found
        :raises AlreadyPairedError: if the alias was already used
        :raises UnavailableError: if the device is already paired
        :raises MaxTriesError: if the device received more than 100 unsuccessful attempts
        :raises BusyError: if a parallel pairing is ongoing
        :raises AuthenticationError: if the verification of the device's SRP proof fails
        :raises MaxPeersError: if the device cannot accept an additional pairing
        :raises UnavailableError: on wrong pin
        :raises MalformedPinError: if the pin is malformed
        """
        Controller.check_pin_format(pin)
        finish_pairing = self.start_pairing(alias, accessory_id)
        return finish_pairing(pin)

    def start_pairing(self, alias, accessory_id):
        """
        This starts a pairing attempt with the IP accessory identified by its id.
        It returns a callable (finish_pairing) which you must call with the pairing pin.

        Accessories can be found via the discover method. The id field is the accessory's id for the second parameter.

        The required pin is either printed on the accessory or displayed. Must be a string of the form 'XXX-YY-ZZZ'. If
        this format is not used, a MalformedPinError is raised.

        Important: no automatic saving of the pairing data is performed. If you don't do this, the information is lost
            and you have to reset the accessory!

        :param alias: the alias for the accessory in the controllers data
        :param accessory_id: the accessory's id
        :param pin: function to return the accessory's pin
        :raises AccessoryNotFoundError: if no accessory with the given id can be found
        :raises AlreadyPairedError: if the alias was already used
        :raises UnavailableError: if the device is already paired
        :raises MaxTriesError: if the device received more than 100 unsuccessful attempts
        :raises BusyError: if a parallel pairing is ongoing
        :raises AuthenticationError: if the verification of the device's SRP proof fails
        :raises MaxPeersError: if the device cannot accept an additional pairing
        :raises UnavailableError: on wrong pin
        """
        if alias in self.pairings:
            raise AlreadyPairedError('Alias "{a}" is already paired.'.format(a=alias))

        connection_data = find_device_ip_and_port(accessory_id)
        if connection_data is None:
            raise AccessoryNotFoundError('Cannot find accessory with id "{i}".'.format(i=accessory_id))
        conn = HomeKitHTTPConnection(connection_data['ip'], port=connection_data['port'])

        try:
            write_fun = create_ip_pair_setup_write(conn)

            state_machine = perform_pair_setup_part1()
            request, expected = state_machine.send(None)
            while True:
                try:
                    response = write_fun(request, expected)
                    request, expected = state_machine.send(response)
                except StopIteration as result:
                    salt, pub_key = result.value
                    break

        except Exception:
            conn.close()
            raise

        def finish_pairing(pin):
            Controller.check_pin_format(pin)
            try:
                state_machine = perform_pair_setup_part2(pin, str(uuid.uuid4()), salt, pub_key)
                request, expected = state_machine.send(None)
                while True:
                    try:
                        response = write_fun(request, expected)
                        request, expected = state_machine.send(response)
                    except StopIteration as result:
                        pairing = result.value
                        break
            finally:
                conn.close()

            pairing['AccessoryIP'] = connection_data['ip']
            pairing['AccessoryPort'] = connection_data['port']
            pairing['Connection'] = 'IP'
            self.pairings[alias] = IpPairing(pairing)

        return finish_pairing

    def perform_pairing_ble(self, alias, accessory_mac, pin, adapter='hci0'):
        """
        This performs a pairing attempt with the Bluetooth LE accessory identified by its mac address.

        Accessories can be found via the discover method. The mac field is the accessory's mac for the second parameter.

        The required pin is either printed on the accessory or displayed. Must be a string of the form 'XXX-YY-ZZZ'. If
        this format is not used, a MalformedPinError is raised.

        Important: no automatic saving of the pairing data is performed. If you don't do this, the information is lost
            and you have to reset the accessory!

        :param alias: the alias for the accessory in the controllers data
        :param accessory_mac: the accessory's mac address
        :param pin: function to return the accessory's pin
        :param adapter: the bluetooth adapter to be used (defaults to hci0)
        :raises MalformedPinError: if the pin is malformed
        # TODO add raised exceptions
        """
        Controller.check_pin_format(pin)
        finish_pairing = self.start_pairing_ble(alias, accessory_mac, adapter)
        return finish_pairing(pin)

    def remove_pairing(self, alias, pairingId=None):
        """
        Remove a pairing between the controller and the accessory. The pairing data is delete on both ends, on the
        accessory and the controller.

        Important: no automatic saving of the pairing data is performed. If you don't do this, the accessory seems still
            to be paired on the next start of the application.

        :param alias: the controller's alias for the accessory
        :param pairingId: the pairing id to be removed
        :raises AuthenticationError: if the controller isn't authenticated to the accessory.
        :raises AccessoryNotFoundError: if the device can not be found via zeroconf
        :raises UnknownError: on unknown errors
        """
        # package visibility like in java would be nice here
        pairing_data = self.pairings[alias]._get_pairing_data()
        connection_type = pairing_data['Connection']
        if not pairingId:
            pairingIdToDelete = pairing_data['iOSPairingId']
        else:
            pairingIdToDelete = pairingId

        # Prepare the common (for IP and BLE) request data
        request_tlv = tlv8.encode([
            tlv8.Entry(TlvTypes.State, States.M1),
            tlv8.Entry(TlvTypes.Method, Methods.RemovePairing),
            tlv8.Entry(TlvTypes.Identifier, pairingIdToDelete.encode())
        ])

        if connection_type == 'IP':
            session = IpSession(pairing_data)
            response = session.post('/pairings', request_tlv, content_type='application/pairing+tlv8')
            session.close()
            data = response.read()
            data = tlv8.decode(data, {
                TlvTypes.State: tlv8.DataType.INTEGER,
                TlvTypes.Error: tlv8.DataType.INTEGER

            })
        elif connection_type == 'BLE':
            raise Exception('not implemented (not IP)')

        # act upon the response
        # handle the result, spec says, if it has only one entry with state == M2 we unpaired, else its an error.
        logging.debug('response data: %s', tlv8.format_string(data))
        state = data.first_by_id(TlvTypes.State).data
        if len(data) == 1 and state == States.M2:
            if not pairingId:
                del self.pairings[alias]
        else:
            error = data.first_by_id(TlvTypes.Error)
            if error and error.data == Errors.Authentication:
                raise AuthenticationError('Remove pairing failed: missing authentication')
            else:
                raise UnknownError('Remove pairing failed: unknown error')
